import numpy as np

from mcdc.constant import INF, GYRATION_RADIUS_ALL, PCT_NONE, PI, SHIFT


class InputDeck:
    def __init__(self):
        self.reset()

    def reset(self):
        self.nuclides = []
        self.materials = []
        self.surfaces = []
        self.cells = []
        self.universes = [{}]
        self.lattices = []
        self.sources = []
        self.meshes = []
        # Default cards are set by functions make_card_*

        # Root universe
        self.universes[0] = make_card_universe(0)
        self.universes[0]["ID"] = 0

        self.tally = {
            "tag": "Tally",
            "tracklength": True,
            "flux": False,
            "density": False,
            "fission": False,
            "total": False,
            "current": False,
            "eddington": False,
            "exit": False,
            "mesh": make_card_mesh(),
        }

        self.setting = {
            "tag": "Setting",
            "mode_MG": True,
            "mode_CE": False,
            "N_particle": 0,
            "N_batch": 1,
            "rng_seed": 1,
            "time_boundary": INF,
            "progress_bar": True,
            "output_name": "output",
            "save_input_deck": True,
            "track_particle": False,
            "mode_eigenvalue": False,
            "k_init": 1.0,
            "N_inactive": 0,
            "N_active": 0,
            "N_cycle": 0,
            "save_particle": False,
            "gyration_radius": False,
            "gyration_radius_type": GYRATION_RADIUS_ALL,
            "N_census": 1,
            "census_time": np.array([INF]),
            "source_file": False,
            "source_file_name": "",
            "IC_file": False,
            "IC_file_name": "",
            "N_precursor": 0,
            # Below are parameters not copied to mcdc.setting
            "bank_active_buff": 100,
            "bank_census_buff": 1.0,
            # TODO: Move to technique
            "N_sensitivity": 0,
        }

        self.technique = {
            "tag": "Technique",
            "weighted_emission": True,
            "implicit_capture": False,
            "population_control": False,
            "pct": PCT_NONE,
            "pc_factor": 1.0,
            "weight_window": False,
            "ww": np.ones([1, 1, 1, 1]),
            "ww_width": 2.5,
            "ww_mesh": make_card_mesh(),
            "weight_roulette": False,
            "wr_threshold": 0.0,
            "wr_chance": 1.0,
            "iQMC": False,
            "iqmc": {
                "generator": "sobol",
                "fixed_source_solver": "source_iteration",
                "eigenmode_solver": "power_iteration",
                "preconditioner_sweeps": 5,
                "krylov_restart": 5,
                "krylov_vector_size": 1,
                "tol": 1e-6,
                "res": 1.0,
                "res_outter": 1.0,
                "itt": 0,
                "itt_outter": 0,
                "maxitt": 5,
                "N_dim": 6,
                "seed": 12345,
                "scramble": False,
                "fixed_source": np.ones([1, 1, 1, 1]),
                "material_idx": np.ones([1, 1, 1, 1]),
                "source": np.ones([1, 1, 1, 1]),
                "score": {
                    "flux": np.ones([1, 1, 1, 1]),
                    "tilt-x": np.zeros([1, 1, 1, 1]),
                    "tilt-y": np.zeros([1, 1, 1, 1]),
                    "tilt-z": np.zeros([1, 1, 1, 1]),
                    "tilt-xy": np.zeros([1, 1, 1, 1]),
                    "tilt-xz": np.zeros([1, 1, 1, 1]),
                    "tilt-yz": np.zeros([1, 1, 1, 1]),
                    "fission-source": np.zeros([1, 1, 1, 1]),
                },
                "score_list": {
                    "flux": True,
                    "effective-scattering": True,
                    "effective-fission": True,
                    "tilt-x": False,
                    "tilt-y": False,
                    "tilt-z": False,
                    "tilt-xy": False,
                    "tilt-xz": False,
                    "tilt-yz": False,
                    "fission-power": False,
                    "fission-source": False,
                },
                "mesh": {
                    "g": np.array([-INF, INF]),
                    "t": np.array([-INF, INF]),
                    "x": np.array([-INF, INF]),
                    "y": np.array([-INF, INF]),
                    "z": np.array([-INF, INF]),
                    "mu": np.array([-1.0, 1.0]),
                    "azi": np.array([-PI, PI]),
                },
            },
            "IC_generator": False,
            "IC_N_neutron": 0,
            "IC_N_precursor": 0,
            "IC_neutron_density": 0.0,
            "IC_precursor_density": 0.0,
            "IC_neutron_density_max": 0.0,
            "IC_precursor_density_max": 0.0,
            "IC_cycle_stretch": 1.0,
            "branchless_collision": False,
            "dsm_order": 1,
            "uq": False,
        }

        self.uq_deltas = {
            "tag": "Uq",
            "nuclides": [],
            "materials": [],
            "surfaces": [],
        }


class SurfaceHandle:
    def __init__(self, card):
        self.card = card

    def __pos__(self):
        return [self.card, True]

    def __neg__(self):
        return [self.card, False]


def make_card_nuclide(G=1, J=0):
    card = {}
    card["tag"] = "Nuclide"
    card["name"] = ""
    card["ID"] = -1
    card["fissionable"] = False
    card["G"] = G
    card["J"] = J
    card["speed"] = np.ones(G)
    card["decay"] = np.ones(J) * INF
    card["capture"] = np.zeros(G)
    card["scatter"] = np.zeros(G)
    card["fission"] = np.zeros(G)
    card["total"] = np.zeros(G)
    card["nu_s"] = np.ones(G)
    card["nu_p"] = np.zeros(G)
    card["nu_d"] = np.zeros([G, J])
    card["nu_f"] = np.zeros(G)
    card["chi_s"] = np.zeros([G, G])
    card["chi_p"] = np.zeros([G, G])
    card["chi_d"] = np.zeros([J, G])
    card["sensitivity"] = False
    card["sensitivity_ID"] = 0
    card["dsm_Np"] = 1.0
    card["uq"] = False
    return card


def make_card_material(N_nuclide, G=1, J=0):
    card = {}
    card["tag"] = "Material"
    card["ID"] = -1
    card["N_nuclide"] = N_nuclide
    card["nuclide_IDs"] = np.zeros(N_nuclide, dtype=int)
    card["nuclide_densities"] = np.zeros(N_nuclide, dtype=float)
    card["G"] = G
    card["J"] = J
    card["speed"] = np.zeros(G)
    card["capture"] = np.zeros(G)
    card["scatter"] = np.zeros(G)
    card["fission"] = np.zeros(G)
    card["total"] = np.zeros(G)
    card["nu_s"] = np.ones(G)
    card["nu_p"] = np.zeros(G)
    card["nu_d"] = np.zeros([G, J])
    card["nu_f"] = np.zeros(G)
    card["chi_s"] = np.zeros([G, G])
    card["chi_p"] = np.zeros([G, G])
    card["name"] = None
    card["sensitivity"] = False
    card["uq"] = False
    return card


def make_card_surface():
    card = {}
    card["tag"] = "Surface"
    card["ID"] = -1
    card["vacuum"] = False
    card["reflective"] = False
    card["A"] = 0.0
    card["B"] = 0.0
    card["C"] = 0.0
    card["D"] = 0.0
    card["E"] = 0.0
    card["F"] = 0.0
    card["G"] = 0.0
    card["H"] = 0.0
    card["I"] = 0.0
    card["J"] = np.array([[0.0, 0.0]])
    card["t"] = np.array([-SHIFT, INF])
    card["N_slice"] = 1
    card["linear"] = False
    card["nx"] = 0.0
    card["ny"] = 0.0
    card["nz"] = 0.0
    card["sensitivity"] = False
    card["sensitivity_ID"] = 0
    card["type"] = " "
    card["dsm_Np"] = 1.0
    return card


def make_card_cell(N_surface):
    card = {}
    card["tag"] = "Cell"
    card["ID"] = -1
    card["N_surface"] = N_surface
    card["surface_IDs"] = np.zeros(N_surface, dtype=int)
    card["positive_flags"] = np.zeros(N_surface, dtype=bool)
    card["material_ID"] = 0
    card["material_name"] = None
    card["lattice"] = False
    card["lattice_ID"] = 0
    card["lattice_center"] = np.array([0.0, 0.0, 0.0])
    return card


def make_card_universe(N_cell):
    card = {}
    card["tag"] = "Universe"
    card["ID"] = -1
    card["N_cell"] = N_cell
    card["cell_IDs"] = np.zeros(N_cell, dtype=int)
    return card


def make_card_lattice():
    card = {}
    card["tag"] = "Lattice"
    card["ID"] = -1
    card["mesh"] = {
        "x0": -INF,
        "dx": 2 * INF,
        "Nx": 1,
        "y0": -INF,
        "dy": 2 * INF,
        "Ny": 1,
        "z0": -INF,
        "dz": 2 * INF,
        "Nz": 1,
    }
    card["universe_IDs"] = np.array([[[[0]]]])
    return card


def make_card_source():
    card = {}
    card["tag"] = "Source"
    card["ID"] = -1
    card["box"] = False
    card["isotropic"] = True
    card["white"] = False
    card["x"] = 0.0
    card["y"] = 0.0
    card["z"] = 0.0
    card["box_x"] = np.array([0.0, 0.0])
    card["box_y"] = np.array([0.0, 0.0])
    card["box_z"] = np.array([0.0, 0.0])
    card["ux"] = 0.0
    card["uy"] = 0.0
    card["uz"] = 0.0
    card["white_x"] = 0.0
    card["white_y"] = 0.0
    card["white_z"] = 0.0
    card["group"] = np.array([1.0])
    card["energy"] = np.array([[14e6, 14e6], [1.0, 1.0]])
    card["time"] = np.array([0.0, 0.0])
    card["prob"] = 1.0
    return card


def make_card_mesh():
    return {
        "x": np.array([-INF, INF]),
        "y": np.array([-INF, INF]),
        "z": np.array([-INF, INF]),
        "t": np.array([-INF, INF]),
        "mu": np.array([-1.0, 1.0]),
        "azi": np.array([-PI, PI]),
        "g": np.array([-INF, INF]),
    }


def make_card_uq():
    return {
        "tag": "t",
        "ID": -1,
        "key": "k",
        "mean": 0.0,
        "delta": 0.0,
        "distribution": "d",
        "rng_seed": 0,
        "group": False,
        "group_group": False,
    }



import numpy as np
import h5py
import sys

sys.path.append("../../")
import tool


# Cases run
N_min = int(sys.argv[1])
N_max = int(sys.argv[2])
N_particle_list = np.logspace(N_min, N_max, (N_max - N_min) * 2 + 1)

# Reference solution
data = np.load("reference.npz")
phi_ref = data["phi"]

# Error containers
error = np.zeros(len(N_particle_list))

error_max = np.zeros(len(N_particle_list))

# Calculate error
for i, N_particle in enumerate(N_particle_list):
    # Get results
    with h5py.File("output_%i.h5" % int(N_particle), "r") as f:
        x = f["tally/grid/x"][:]
        dx = x[1:] - x[:-1]
        t = f["tally/grid/t"][:]
        dt = t[1:] - t[:-1]
        K = len(t) - 1

        phi = f["tally/flux/mean"][:]

    # Normalize
    for k in range(K):
        phi[k] /= dx * dt[k]

    # Get error
    error[i] = tool.error(phi, phi_ref)

    error_max[i] = tool.error_max(phi, phi_ref)

# Plot
tool.plot_convergence("azurv1_pl_flux", N_particle_list, error, error_max)



import numpy as np
from scipy.integrate import quad


def reference():
    # Load material data
    with np.load("SHEM-361.npz") as data:
        SigmaT = data["SigmaT"]
        SigmaC = data["SigmaC"]
        SigmaS = data["SigmaS"]
        nuSigmaF = data["nuSigmaF"]
        G = data["G"]
    SigmaT += SigmaC * 0.5

    A = np.diag(SigmaT) - SigmaS - nuSigmaF
    Q = np.zeros(G)
    Q[-1] = 1.0

    phi = np.linalg.solve(A, Q)
    return phi



from reference import reference
import numpy as np
import h5py
import sys

sys.path.append("../../")
import tool


# Cases run
N_min = int(sys.argv[1])
N_max = int(sys.argv[2])
N_particle_list = np.logspace(N_min, N_max, (N_max - N_min) * 2 + 1)

# Reference solution
with h5py.File("output_%i.h5" % int(N_particle_list[0]), "r") as f:
    x = f["tally/grid/x"][:]
    dx = x[1:] - x[:-1]
phi_ref = reference(x)

# Error containers
error = np.zeros(len(N_particle_list))

error_max = np.zeros(len(N_particle_list))

# Calculate error
for k, N_particle in enumerate(N_particle_list):
    # Get results
    with h5py.File("output_%i.h5" % int(N_particle), "r") as f:
        phi = f["tally/flux/mean"][:] / dx * 101.0

    # Get error
    error[k] = tool.error(phi, phi_ref)
    error_max[k] = tool.error_max(phi, phi_ref)

tool.plot_convergence("reed_flux", N_particle_list, error, error_max)



import matplotlib.pyplot as plt
import numpy as np


def error(val, ref):
    return np.sqrt(np.average((val - ref) ** 2) / np.sum(ref**2))


def error_max(val, ref):
    return np.max(val - ref) / np.max(ref)


def rerror(val, ref):
    return np.sqrt(np.sum(((val - ref) / ref) ** 2)) / ref.size


def rerror_max(val, ref):
    return np.max(np.abs((val - ref) / ref))


def plot_convergence(name, N_particle, error, error_max):
    mid = int(len(N_particle) / 2)

    plt.plot(N_particle, error, "bo", fillstyle="none", label="2-norm")

    plt.plot(N_particle, error_max, "gD", fillstyle="none", label="max")

    line = 1.0 / np.sqrt(N_particle)
    line *= error[mid] / line[mid]
    plt.plot(N_particle, line, "r--", label=r"$O(N^{-0.5})$")

    line = 1.0 / np.sqrt(N_particle)
    line *= error_max[mid] / line[mid]
    plt.plot(N_particle, line, "r--")

    plt.xscale("log")
    plt.yscale("log")
    plt.ylabel("Relative error")
    plt.xlabel(r"# of histories, $N$")
    plt.legend()
    plt.grid()
    plt.title(name)
    plt.savefig("../../results/" + name + ".png")
    # plt.show()
    plt.clf()



import numpy as np
import matplotlib.pyplot as plt
import h5py
from matplotlib import cm
from matplotlib import colors


# =============================================================================
# Plot results
# =============================================================================

# Get results
with h5py.File("output.h5", "r") as f:
    fis_avg = f["tally/fission/mean"][:]
    fis_sd = f["tally/fission/sdev"][:]
    t = f["tally/grid/t"][:]
t_mid = 0.5 * (t[:-1] + t[1:])
dt = t[1:] - t[:-1]

# Get reference
with h5py.File("reference.h5", "r") as f:
    fis_ref = f["tally/fission/mean"][:]
    fis_ref_sd = f["tally/fission/sdev"][:]

# Normalize
norm = fis_ref[0] / dt[0]
fis_avg /= norm * dt
fis_sd /= norm * dt
fis_ref /= norm * dt
fis_ref_sd /= norm * dt

# Plot
plt.plot(t_mid, fis_ref, "-m", fillstyle="none", label="Ref. (MC)")
plt.fill_between(
    t_mid, fis_ref - fis_ref_sd, fis_ref + fis_ref_sd, alpha=0.2, color="m"
)
plt.plot(t_mid, fis_avg, "ok", fillstyle="none", label="MC")
plt.fill_between(t_mid, fis_avg - fis_sd, fis_avg + fis_sd, alpha=0.2, color="k")
plt.yscale("log")
plt.ylabel("Normalized fission rate")
plt.xlabel("time [s]")
plt.axvspan(0, 5, facecolor="gray", alpha=0.2)
plt.axvspan(5, 10, facecolor="green", alpha=0.2)
plt.axvspan(10, 15, facecolor="red", alpha=0.2)
plt.axvspan(15, 20, facecolor="blue", alpha=0.2)
plt.annotate(
    "Phase 1",
    (2.5, 0.3),
    color="black",
    ha="center",
    va="center",
    backgroundcolor="white",
)
plt.annotate(
    "Phase 2",
    (7.5, 0.3),
    color="black",
    ha="center",
    va="center",
    backgroundcolor="white",
)
plt.annotate(
    "Phase 3",
    (12.5, 0.3),
    color="black",
    ha="center",
    va="center",
    backgroundcolor="white",
)
plt.annotate(
    "Phase 4",
    (17.5, 4.0),
    color="black",
    ha="center",
    va="center",
    backgroundcolor="white",
)
plt.xlim([0.0, 20.0])
plt.ylim([0.09, 200.0])
plt.grid(which="both")
plt.legend()
plt.show()



import numba as nb
import numpy as np
import sys
from mpi4py import MPI


master = MPI.COMM_WORLD.Get_rank() == 0


def print_msg(msg):
    if master:
        print(msg)
        sys.stdout.flush()


def print_error(msg):
    print("ERROR: %s\n" % msg)
    sys.stdout.flush()
    sys.exit()


def print_warning(msg):
    if master:
        print("Warning: %s\n" % msg)
        sys.stdout.flush()


def print_banner(mcdc):
    size = MPI.COMM_WORLD.Get_size()
    if master:
        banner = (
            "\n"
            + r"  __  __  ____  __ ____   ____ "
            + "\n"
            + r" |  \/  |/ ___|/ /_  _ \ / ___|"
            + "\n"
            + r" | |\/| | |   /_  / | | | |    "
            + "\n"
            + r" | |  | | |___ / /| |_| | |___ "
            + "\n"
            + r" |_|  |_|\____|// |____/ \____|"
            + "\n"
            + "\n"
        )
        if nb.config.DISABLE_JIT:
            banner += "           Mode | Python\n"
        else:
            banner += "           Mode | Numba\n"
        if mcdc["technique"]["iQMC"]:
            banner += "      Algorithm | iQMC\n"
            rng = mcdc["technique"]["iqmc"]["generator"]
            if mcdc["setting"]["mode_eigenvalue"]:
                solver = mcdc["technique"]["iqmc"]["eigenmode_solver"]
            else:
                solver = mcdc["technique"]["iqmc"]["fixed_source_solver"]
            banner += "            RNG | " + rng + "\n"
            banner += "         Solver | " + solver + "\n"
        else:
            banner += "      Algorithm | History-based\n"
        banner += "  MPI Processes | %i\n" % size
        banner += " OpenMP Threads | 1"
        print(banner)
        sys.stdout.flush()


def print_progress(percent, mcdc):
    if master:
        sys.stdout.write("\r")
        if not mcdc["setting"]["mode_eigenvalue"]:
            if mcdc["setting"]["N_census"] == 1:
                sys.stdout.write(
                    " [%-28s] %d%%" % ("=" * int(percent * 28), percent * 100.0)
                )
            else:
                idx = mcdc["idx_census"] + 1
                N = len(mcdc["setting"]["census_time"])
                sys.stdout.write(
                    " Census %i/%i: [%-28s] %d%%"
                    % (idx, N, "=" * int(percent * 28), percent * 100.0)
                )
        else:
            if mcdc["setting"]["gyration_radius"]:
                sys.stdout.write(
                    " [%-40s] %d%%" % ("=" * int(percent * 40), percent * 100.0)
                )
            else:
                sys.stdout.write(
                    "[%-32s] %d%%" % ("=" * int(percent * 32), percent * 100.0)
                )
        sys.stdout.flush()


def print_progress_iqmc(mcdc):
    # TODO: function was not working with numba when structured like the
    # other print_progress functions
    if master:
        if mcdc["setting"]["progress_bar"]:
            sys.stdout.write("\r")
            itt = mcdc["technique"]["iqmc"]["itt"]
            res = mcdc["technique"]["iqmc"]["res"]
            print("\n*******************************")
            print("Iteration  %2d" % (itt))
            print("Residual %10.3E" % (res))
            print("*******************************\n")
            sys.stdout.flush()


def print_header_eigenvalue(mcdc):
    if master:
        if mcdc["setting"]["gyration_radius"]:
            print("\n #     k        GyRad.  k (avg)            ")
            print(" ====  =======  ======  ===================")
        elif mcdc["technique"]["iQMC"]:
            print("\n #     k        Residual         ")
            print(" ==== ======= ===================")
        else:
            print("\n #     k        k (avg)            ")
            print(" ====  =======  ===================")


def print_header_batch(mcdc):
    idx_batch = mcdc["idx_batch"]
    if master:
        print("\nBatch %i/%i" % (idx_batch + 1, mcdc["setting"]["N_batch"]))


def print_progress_eigenvalue(mcdc):
    if master:
        idx_cycle = mcdc["idx_cycle"]
        k_eff = mcdc["k_eff"]
        k_avg = mcdc["k_avg_running"]
        k_sdv = mcdc["k_sdv_running"]
        gr = mcdc["gyration_radius"][idx_cycle]
        if mcdc["setting"]["progress_bar"]:
            sys.stdout.write("\r")
            sys.stdout.write("\033[K")
        if mcdc["setting"]["gyration_radius"]:
            if not mcdc["cycle_active"]:
                print(" %-4i  %.5f  %6.2f" % (idx_cycle + 1, k_eff, gr))
            else:
                print(
                    " %-4i  %.5f  %6.2f  %.5f +/- %.5f"
                    % (idx_cycle + 1, k_eff, gr, k_avg, k_sdv)
                )
        else:
            if not mcdc["cycle_active"]:
                print(" %-4i  %.5f" % (idx_cycle + 1, k_eff))
            else:
                print(
                    " %-4i  %.5f  %.5f +/- %.5f" % (idx_cycle + 1, k_eff, k_avg, k_sdv)
                )


def print_iqmc_eigenvalue_progress(mcdc):
    if master:
        if mcdc["setting"]["progress_bar"]:
            sys.stdout.write("\r")
            k_eff = mcdc["k_eff"]
            itt = mcdc["technique"]["iqmc"]["itt_outter"]
            res = mcdc["technique"]["iqmc"]["res_outter"]
            print("\n %2d   %2.5f  %10.3E" % (itt, k_eff, res))
            sys.stdout.flush()


def print_iqmc_eigenvalue_exit_code(mcdc):
    if master:
        if mcdc["setting"]["progress_bar"]:
            sys.stdout.write("\r")
            maxit = mcdc["technique"]["iqmc"]["maxitt"]
            itt = mcdc["technique"]["iqmc"]["itt_outter"]
            if itt >= maxit:
                print("\n")
                print("================================")
                print("\n")
                print(
                    " Convergence to tolerance not achieved: Maximum number of iterations."
                )
            else:
                print("\n")
                print("================================")
                print(" Successful convergence.")
                print("\n")
            sys.stdout.flush()


def print_runtime(mcdc):
    total = mcdc["runtime_total"]
    preparation = mcdc["runtime_preparation"]
    simulation = mcdc["runtime_simulation"]
    output = mcdc["runtime_output"]
    if master:
        print("\n Runtime report:")
        print_time("Total      ", total, 100)
        print_time("Preparation", preparation, preparation / total * 100)
        print_time("Simulation ", simulation, simulation / total * 100)
        print_time("Output     ", output, output / total * 100)
        print("\n")
        sys.stdout.flush()


def print_time(tag, t, percent):
    if t >= 24 * 60 * 60:
        print("   %s | %.2f days (%.1f%%)" % (tag, t / 24 / 60 / 60), percent)
    elif t >= 60 * 60:
        print("   %s | %.2f hours (%.1f%%)" % (tag, t / 60 / 60, percent))
    elif t >= 60:
        print("   %s | %.2f minutes (%.1f%%)" % (tag, t / 60, percent))
    else:
        print("   %s | %.2f seconds (%.1f%%)" % (tag, t, percent))


def print_bank(bank, show_content=False):
    tag = bank["tag"]
    size = bank["size"]
    particles = bank["particles"]

    print("\n=============")
    print("Particle bank")
    print("  tag  :", tag)
    print("  size :", size, "of", len(bank["particles"]))
    if show_content and size > 0:
        for i in range(size):
            print(" ", particles[i])
    print("\n")



import numpy as np
import matplotlib.pyplot as plt
import matplotlib.animation as animation
import h5py
import sys


# Reference solution
data = np.load("reference.npz")
phi_ref = data["phi"]

# Get results
output = sys.argv[1]
with h5py.File(output, "r") as f:
    x = f["tally/grid/x"][:]
    dx = x[1:] - x[:-1]
    x_mid = 0.5 * (x[:-1] + x[1:])
    t = f["tally/grid/t"][:]
    dt = t[1:] - t[:-1]
    K = len(t) - 1

    phi = f["tally/flux/mean"][:]
    phi_sd = f["tally/flux/sdev"][:]

    # Normalize
    for k in range(K):
        phi[k] /= dx * dt[k]
        phi_sd[k] /= dx * dt[k]

# Flux - average
fig = plt.figure()
ax = plt.axes(
    xlim=(-21.889999999999997, 21.89), ylim=(-0.042992644459595206, 0.9028455336514992)
)
ax.grid()
ax.set_xlabel(r"$x$")
ax.set_ylabel(r"Flux")
ax.set_title(r"$\bar{\phi}_{k,j}$")
(line1,) = ax.plot([], [], "-b", label="MC")
(line2,) = ax.plot([], [], "--r", label="Ref.")
fb = ax.fill_between([], [], [], [], alpha=0.2, color="b")
text = ax.text(0.02, 0.9, "", transform=ax.transAxes)
ax.legend()


def animate(k):
    global fb
    fb.remove()
    line1.set_data(x_mid, phi[k, :])
    fb = ax.fill_between(
        x_mid, phi[k, :] - phi_sd[k, :], phi[k, :] + phi_sd[k, :], alpha=0.2, color="b"
    )
    line2.set_data(x_mid, phi_ref[k, :])
    text.set_text(r"$t \in [%.1f,%.1f]$ s" % (t[k], t[k + 1]))
    return line1, line2, text


simulation = animation.FuncAnimation(fig, animate, frames=K)
writervideo = animation.FFMpegWriter(fps=6)
plt.show()



import openmc

rho_uo2 = 10.97
rho_h2o = 0.997
rho_b4c = 2.52

a_u235 = 0.05
a_u238 = 1.0 - a_u235

A_uo2 = 270.03
A_h2o = 18.01528
A_b4c = 55.255

N_avo = 6.023e23

N_uo2 = rho_uo2 * N_avo / A_uo2 * 1e-24
N_h2o = rho_h2o * N_avo / A_h2o * 1e-24
N_b4c = rho_b4c * N_avo / A_b4c * 1e-24

N_u235 = a_u235 * N_uo2
N_u238 = a_u238 * N_uo2
N_o16_uo2 = 2.0 * N_uo2

N_mix = N_u235 + N_u238 + N_o16_uo2
N_u235 = N_u235 / N_mix
N_u238 = N_u238 / N_mix
N_o16_uo2 = N_o16_uo2 / N_mix

uo2 = openmc.Material()
uo2.set_density("atom/b-cm", N_mix)
uo2.add_nuclide("U235", N_u235)
uo2.add_nuclide("U238", N_u238)
uo2.add_nuclide("O16", N_o16_uo2)

N_h1 = 2.0 * N_h2o
N_o16_h2o = N_h2o

N_mix = N_h1 + N_o16_h2o
N_h1 = N_h1 / N_mix
N_o16_h2o = N_o16_h2o / N_mix

h2o = openmc.Material()
h2o.set_density("atom/b-cm", N_mix)
h20.add_nuclide("H1", N_h1)
h20.add_nuclide("O16", N_o16_h2o)

N_b10 = 4.0 * N_b4c
N_c12 = N_b4c

N_mix = N_b10 + N_c12
N_b10 = N_b10 / N_mix
N_c12 = N_c12 / N_mix

b4c = openmc.Material()
b4c.set_density("atom/b-cm", N_mix)
b4c.add_nuclide("B10", N_b10)
b4c.add_nuclide("C12", N_c12)

mats = openmc.Materials([h20, b4c, uo2])
mats.export_to_xml()

# Create a 5 cm x 5 cm box filled with iron
box = openmc.model.rectangular_prism(10.0, 10.0, boundary_type="vacuum")
cell = openmc.Cell(fill=iron, region=box)
geometry = openmc.Geometry([cell])
geometry.export_to_xml()

# Tell OpenMC we're going to use our custom source
settings = openmc.Settings()
settings.run_mode = "fixed source"
settings.batches = 10
settings.particles = 1000
source = openmc.CompiledSource()
source.library = "build/libsource.so"
settings.source = source
settings.export_to_xml()

# Finally, define a mesh tally so that we can see the resulting flux
mesh = openmc.RegularMesh()
mesh.lower_left = (-5.0, -5.0)
mesh.upper_right = (5.0, 5.0)
mesh.dimension = (50, 50)

tally = openmc.Tally()
tally.filters = [openmc.MeshFilter(mesh)]
tally.scores = ["flux"]
tallies = openmc.Tallies([tally])
tallies.export_to_xml()



import numpy as np

from mcdc import type_
from mcdc.kernel import surface_evaluate
from mcdc.constant import INF


# =============================================================================
# Static surface
# =============================================================================


def tmp_plane_x():
    type_.make_type_surface(1)
    P = np.zeros(1, dtype=type_.particle)[0]
    S = np.zeros(1, dtype=type_.surface)[0]

    x = 5.0
    trans = np.array([0.0, 0.0, 0.0])

    S["G"] = 1.0
    S["linear"] = True
    S["J"] = np.array([-x, -x])
    S["t"] = np.array([0.0, INF])

    # Surface on the left
    P["x"] = 4.0
    result = surface_evaluate(P, S, trans)
    assert result < 0.0

    # Surface on the right
    P["x"] = 9.0
    result = surface_evaluate(P, S, trans)
    assert result > 0.0



import numpy as np
import os
import sys

if len(sys.argv) > 1:
    N_proc = int(sys.argv[1])
else:
    N_proc = 1


waitlist = []

# Fixed source
N_min = 3
N_max = 7
for task in os.scandir("./fixed_source"):
    os.chdir(task)
    for N_hist in np.logspace(N_min, N_max, (N_max - N_min) * 2 + 1):
        if not os.path.isfile("output_" + str(int(N_hist)) + ".h5"):
            print(task, int(N_hist))
            if N_proc == 1:
                os.system(
                    "python input.py --mode=numba --N_particle=%i --output=output_%i"
                    % (N_hist, N_hist)
                )
            else:
                os.system(
                    "srun -n %i python input.py --mode=numba --N_particle=%i --output=output_%i"
                    % (N_proc, N_hist, N_hist)
                )
    os.chdir(r"../..")

# Eigenvalue
"""
N_min = 1
N_max = 3
for task in os.scandir('./eigenvalue'):
    os.chdir(task)
    for N_hist in np.logspace(N_min, N_max, (N_max-N_min)*4+1):
        print(task, int(N_hist))
        if N_proc == 1:
            os.system("python input.py --mode=numba %i"%(N_hist))
        else:
            os.system("srun -n %i python input.py --mode=numba %i"%(N_proc,N_hist))
    os.chdir(r"../..")
"""



import numpy as np
import mcdc


# =============================================================================
# Set model
# =============================================================================
# A shielding problem based on Problem 2 of [Coper NSE 2001]
# https://ans.tandfonline.com/action/showCitFormats?doi=10.13182/NSE00-34

# Set materials
SigmaT = 5.0
c = 0.8
m_barrier = mcdc.material(capture=np.array([SigmaT]), scatter=np.array([[SigmaT * c]]))
SigmaT = 1.0
m_room = mcdc.material(capture=np.array([SigmaT]), scatter=np.array([[SigmaT * c]]))

# Set surfaces
sx1 = mcdc.surface("plane-x", x=0.0, bc="reflective")
sx2 = mcdc.surface("plane-x", x=2.0)
sx3 = mcdc.surface("plane-x", x=2.4)
sx4 = mcdc.surface("plane-x", x=4.0, bc="vacuum")
sy1 = mcdc.surface("plane-y", y=0.0, bc="reflective")
sy2 = mcdc.surface("plane-y", y=2.0)
sy3 = mcdc.surface("plane-y", y=4.0, bc="vacuum")

# Set cells
mcdc.cell([+sx1, -sx2, +sy1, -sy2], m_room)
mcdc.cell([+sx1, -sx4, +sy2, -sy3], m_room)
mcdc.cell([+sx3, -sx4, +sy1, -sy2], m_room)
mcdc.cell([+sx2, -sx3, +sy1, -sy2], m_barrier)

# =============================================================================
# iQMC Parameters
# =============================================================================
N = 1e4
Nx = Ny = 40
maxit = 30
tol = 1e-6
x = np.linspace(0, 4, num=Nx + 1)
y = np.linspace(0, 4, num=Ny + 1)
generator = "halton"
solver = "gmres"

# fixed source in lower left corner
fixed_source = np.zeros((Nx, Ny))
fixed_source[0 : int(0.25 * Nx), 0 : int(0.25 * Nx)] = 1

phi0 = np.ones((Nx, Ny))

mcdc.iQMC(
    x=x,
    y=y,
    fixed_source=fixed_source,
    phi0=phi0,
    maxitt=maxit,
    tol=tol,
    generator=generator,
    fixed_source_solver=solver,
    score=["tilt-x", "tilt-y"],
)

# =============================================================================
# Set tally, setting, and run mcdc
# =============================================================================
# Setting
mcdc.setting(N_particle=N)
# Run
mcdc.run()



import numpy as np
import matplotlib.pyplot as plt
import h5py
from matplotlib import cm
from matplotlib import colors


# =============================================================================
# Plot results
# =============================================================================

# Results
with h5py.File("output.h5", "r") as f:
    phi_avg = f["tally/flux/mean"][:]
    phi_sd = f["tally/flux/sdev"][:]
    fis_avg = f["tally/fission/mean"][:]
    fis_sd = f["tally/fission/sdev"][:]
    k = f["k_cycle"][:]
    k_avg = f["k_mean"][()]
    k_sd = f["k_sdev"][()]
    rg = f["gyration_radius"][:]
    x = f["tally/grid/x"][:]
    y = f["tally/grid/y"][:]
    z = f["tally/grid/z"][:]

dx = x[1] - x[0]
dz = z[1] - z[0]
dV = dx * dx * dz

norm = np.sum(phi_avg)
phi_avg = phi_avg / norm / dV
phi_sd = phi_sd / norm / dV
fis_avg = fis_avg / norm / dV
fis_sd = fis_sd / norm / dV

phi_fast = np.zeros_like(phi_avg[0])
phi_thermal = np.zeros_like(phi_avg[0])
phi_fast_sd = np.zeros_like(phi_avg[0])
phi_thermal_sd = np.zeros_like(phi_avg[0])
fis_ = np.zeros_like(phi_avg[0])
fis__sd = np.zeros_like(phi_avg[0])

for i in range(2):
    phi_fast += phi_avg[i]
    phi_fast_sd += np.square(phi_sd[i])
phi_fast_sd = np.sqrt(phi_fast_sd)
for i in range(2, 7):
    phi_thermal += phi_avg[i]
    phi_thermal_sd += np.square(phi_sd[i])
phi_thermal_sd = np.sqrt(phi_thermal_sd)
for i in range(7):
    fis_ += fis_avg[i]
    fis__sd += np.square(fis_sd[i])
fis__sd = np.sqrt(fis__sd)

print("k = %.5f +- %.5f" % (k_avg, k_sd))

# Plot
N_iter = len(k)
(p1,) = plt.plot(np.arange(1, N_iter + 1), k, "-b", label="MC")
(p2,) = plt.plot(
    np.arange(1, N_iter + 1), np.ones(N_iter) * k_avg, ":r", label="MC-avg"
)
plt.fill_between(
    np.arange(1, N_iter + 1),
    np.ones(N_iter) * (k_avg - k_sd),
    np.ones(N_iter) * (k_avg + k_sd),
    alpha=0.2,
    color="r",
)
plt.xlabel("Iteration #")
plt.ylabel(r"$k$")
plt.grid()
ax2 = plt.gca().twinx()
(p3,) = ax2.plot(np.arange(1, N_iter + 1), rg, "g--", label="GyRad")
plt.ylabel(r"Gyration radius [cm]")
lines = [p1, p2, p3]
plt.legend(lines, [l.get_label() for l in lines])
plt.show()

# X-Z plane
x_mid = 0.5 * (x[1:] + x[:-1])
z_mid = 0.5 * (z[1:] + z[:-1])
X, Z = np.meshgrid(z_mid, x_mid)
phi_fast_xz = np.sum(phi_fast, axis=1)
phi_thermal_xz = np.sum(phi_thermal, axis=1)
fis_xz = np.sum(fis_, axis=1)
phi_fast_xz_sd = np.sqrt(np.sum(np.square(phi_fast_sd), axis=1)) / phi_fast_xz
phi_thermal_xz_sd = np.sqrt(np.sum(np.square(phi_thermal_sd), axis=1)) / phi_thermal_xz
fis_xz_sd = np.sqrt(np.sum(np.square(fis__sd), axis=1)) / fis_xz

plt.pcolormesh(Z, X, phi_fast_xz, shading="nearest")
plt.colorbar()
ax = plt.gca()
ax.set_aspect("equal")
plt.xlabel(r"$x$ [cm]")
plt.ylabel(r"$z$ [cm]")
plt.title(r"Fast neutron flux")
plt.show()

plt.pcolormesh(Z, X, phi_fast_xz_sd, shading="nearest")
plt.colorbar()
ax = plt.gca()
ax.set_aspect("equal")
plt.xlabel(r"$x$ [cm]")
plt.ylabel(r"$z$ [cm]")
plt.title(r"Fast neutron flux stdev")
plt.show()

plt.pcolormesh(Z, X, phi_thermal_xz, shading="nearest")
plt.colorbar()
ax = plt.gca()
ax.set_aspect("equal")
plt.xlabel(r"$x$ [cm]")
plt.ylabel(r"$z$ [cm]")
plt.title(r"Thermal neutron flux")
plt.show()

plt.pcolormesh(Z, X, phi_thermal_xz_sd, shading="nearest")
plt.colorbar()
ax = plt.gca()
ax.set_aspect("equal")
plt.xlabel(r"$x$ [cm]")
plt.ylabel(r"$z$ [cm]")
plt.title(r"Thermal neutron flux stdev")
plt.show()

plt.pcolormesh(Z, X, fis_xz, shading="nearest")
plt.colorbar()
ax = plt.gca()
ax.set_aspect("equal")
plt.xlabel(r"$x$ [cm]")
plt.ylabel(r"$z$ [cm]")
plt.title(r"Fission rate")
plt.show()

plt.pcolormesh(Z, X, fis_xz_sd, shading="nearest")
plt.colorbar()
ax = plt.gca()
ax.set_aspect("equal")
plt.xlabel(r"$x$ [cm]")
plt.ylabel(r"$z$ [cm]")
plt.title(r"Fission rate stdev")
plt.show()

# X-Y plane
Y, X = np.meshgrid(x_mid, x_mid)
phi_fast_xy = np.sum(phi_fast, axis=2)
phi_thermal_xy = np.sum(phi_thermal, axis=2)
fis_xy = np.sum(fis_, axis=2)
phi_fast_xy_sd = np.sqrt(np.sum(np.square(phi_fast_sd), axis=2)) / phi_fast_xy
phi_thermal_xy_sd = np.sqrt(np.sum(np.square(phi_thermal_sd), axis=2)) / phi_thermal_xy
fis_xy_sd = np.sqrt(np.sum(np.square(fis__sd), axis=2)) / fis_xy

plt.pcolormesh(X, Y, phi_fast_xy, shading="nearest")
plt.colorbar()
ax = plt.gca()
ax.set_aspect("equal")
plt.xlabel(r"$x$ [cm]")
plt.ylabel(r"$y$ [cm]")
plt.title(r"Fast neutron flux")
plt.show()

plt.pcolormesh(X, Y, phi_fast_xy_sd, shading="nearest")
plt.colorbar()
ax = plt.gca()
ax.set_aspect("equal")
plt.xlabel(r"$x$ [cm]")
plt.ylabel(r"$y$ [cm]")
plt.title(r"Fast neutron flux stdev")
plt.show()

plt.pcolormesh(X, Y, phi_thermal_xy, shading="nearest")
plt.colorbar()
ax = plt.gca()
ax.set_aspect("equal")
plt.xlabel(r"$x$ [cm]")
plt.ylabel(r"$y$ [cm]")
plt.title(r"Thermal neutron flux")
plt.show()

plt.pcolormesh(X, Y, phi_thermal_xy_sd, shading="nearest")
plt.colorbar()
ax = plt.gca()
ax.set_aspect("equal")
plt.xlabel(r"$x$ [cm]")
plt.ylabel(r"$y$ [cm]")
plt.title(r"Thermal neutron flux stdev")
plt.show()

plt.pcolormesh(X, Y, fis_xy, shading="nearest")
plt.colorbar()
ax = plt.gca()
ax.set_aspect("equal")
plt.xlabel(r"$x$ [cm]")
plt.ylabel(r"$y$ [cm]")
plt.title(r"Fission rate")
plt.show()

plt.pcolormesh(X, Y, fis_xy_sd, shading="nearest")
plt.colorbar()
ax = plt.gca()
ax.set_aspect("equal")
plt.xlabel(r"$x$ [cm]")
plt.ylabel(r"$y$ [cm]")
plt.title(r"Fission rate stdev")
plt.show()


"""
# Mesh grid
X,Y,Z = np.meshgrid(x,y,z)

# Voxel fill
fill = np.zeros_like(phi_fast, dtype=bool)
Nx = len(x)-1
Ny = len(y)-1
Nz = len(z)-1

fill[int(Nx/2):,int(Ny/2):,:] = True

# Convert data to colors
norm = colors.Normalize()
phi_fast = cm.viridis(norm(phi_fast))

ax = plt.figure().add_subplot(projection='3d')
ax.voxels(X, Y, Z, fill, facecolors=phi_fast)
ax.set(xlabel='x (cm)', ylabel='y (cm)', zlabel='z (cm)')
plt.show()
"""



import numpy as np
from scipy.integrate import quad


def reference(x, mu):
    dx = x[1:] - x[:-1]
    dmu = mu[1:] - mu[:-1]
    I = len(x) - 1
    N = len(mu) - 1

    # Parameters
    SigmaT3 = 1.0
    SigmaT1 = 1.5
    SigmaT2 = 2.0
    q1 = 1.0 / 6.0 / 2
    q2 = 1.0 / 6.0 / 2
    q3 = 1.0 / 6.0 / 2
    x1 = 2.0
    x2 = 4.0
    x3 = 6.0
    tau1 = SigmaT1 * x1
    tau2 = SigmaT2 * (x2 - x1)
    tau3 = SigmaT3 * (x3 - x2)

    # Angular flux
    def psi1(mu, x):
        if mu > 0.0:
            return q1 / SigmaT1 * (1.0 - np.exp(-SigmaT1 * x / mu))
        elif mu < 0.0:
            return (psi2(mu, x1) - q1 / SigmaT1) * np.exp(
                -SigmaT1 * (x1 - x) / np.abs(mu)
            ) + q1 / SigmaT1

    def psi2(mu, x):
        if mu > 0.0:
            return (psi1(mu, x1) - q2 / SigmaT2) * np.exp(
                -SigmaT2 * (x - x1) / mu
            ) + q2 / SigmaT2
        elif mu < 0.0:
            return (psi3(mu, x2) - q2 / SigmaT2) * np.exp(
                -SigmaT2 * (x2 - x) / np.abs(mu)
            ) + q2 / SigmaT2

    def psi3(mu, x):
        if mu > 0.0:
            return (psi2(mu, x2) - q3 / SigmaT3) * np.exp(
                -SigmaT3 * (x - x2) / mu
            ) + q3 / SigmaT3
        elif mu < 0.0:
            return q3 / SigmaT3 * (1.0 - np.exp(-SigmaT3 * (x3 - x) / np.abs(mu)))

    # Flux
    def phi1(x):
        return quad(psi1, -1, 1, args=(x), points=[0.0])[0]

    def phi2(x):
        return quad(psi2, -1, 1, args=(x), points=[0.0])[0]

    def phi3(x):
        return quad(psi3, -1, 1, args=(x), points=[0.0])[0]

    # Integrands for current
    def mu_psi1(mu, x):
        if mu > 0.0:
            return mu * psi1(mu, x)
        elif mu < 0.0:
            return mu * psi1(mu, x)

    def mu_psi2(mu, x):
        if mu > 0.0:
            return mu * psi2(mu, x)
        elif mu < 0.0:
            return mu * psi2(mu, x)

    def mu_psi3(mu, x):
        if mu > 0.0:
            return mu * psi3(mu, x)
        elif mu < 0.0:
            return mu * psi3(mu, x)

    # Current
    def J1(x):
        return quad(mu_psi1, -1, 1, args=(x), points=[0.0])[0]

    def J2(x):
        return quad(mu_psi2, -1, 1, args=(x), points=[0.0])[0]

    def J3(x):
        return quad(mu_psi3, -1, 1, args=(x), points=[0.0])[0]

    # Angular flux
    def psi1_(x, mu0, mu1):
        return quad(psi1, mu0, mu1, args=(x), points=[0.0])[0]

    def psi2_(x, mu0, mu1):
        return quad(psi2, mu0, mu1, args=(x), points=[0.0])[0]

    def psi3_(x, mu0, mu1):
        return quad(psi3, mu0, mu1, args=(x), points=[0.0])[0]

    phi = np.zeros(I)
    psi = np.zeros((I, N))
    J = np.zeros(I)

    for i in range(int(I / 3)):
        phi[i] = quad(phi1, x[i], x[i + 1])[0] / dx[i]
        J[i] = quad(J1, x[i], x[i + 1])[0] / dx[i]
        for n in range(N):
            mu0 = mu[n]
            mu1 = mu[n + 1]
            psi[i, n] = quad(psi1_, x[i], x[i + 1], args=(mu0, mu1))[0] / dx[i] / dmu[n]
    for i in range(int(I / 3), int(2 * I / 3)):
        phi[i] = quad(phi2, x[i], x[i + 1])[0] / dx[i]
        J[i] = quad(J2, x[i], x[i + 1])[0] / dx[i]
        for n in range(N):
            mu0 = mu[n]
            mu1 = mu[n + 1]
            psi[i, n] = quad(psi2_, x[i], x[i + 1], args=(mu0, mu1))[0] / dx[i] / dmu[n]
    for i in range(int(2 * I / 3), I):
        phi[i] = quad(phi3, x[i], x[i + 1])[0] / dx[i]
        J[i] = quad(J3, x[i], x[i + 1])[0] / dx[i]
        for n in range(N):
            mu0 = mu[n]
            mu1 = mu[n + 1]
            psi[i, n] = quad(psi3_, x[i], x[i + 1], args=(mu0, mu1))[0] / dx[i] / dmu[n]
    for n in range(N):
        mu0 = mu[n]
        mu1 = mu[n + 1]

    return phi, J, psi



import numpy as np
import matplotlib.pyplot as plt
import h5py
import sys
import matplotlib.animation as animation

# Reference solution
data = np.load("reference.npz")
phi_ref = data["phi"].T
n_ref = data["n"]

# Load results
output = sys.argv[1]
with np.load("SHEM-361.npz") as data:
    E = data["E"]
    G = data["G"]
    speed = data["v"]
    E_mid = 0.5 * (E[1:] + E[:-1])
    dE = E[1:] - E[:-1]
with h5py.File(output, "r") as f:
    t = f["tally/grid/t"][:]
    dt = t[1:] - t[:-1]
    t_mid = 0.5 * (t[1:] + t[:-1])
    K = len(t) - 1

    phi = f["tally/flux/mean"][:]
    phi_sd = f["tally/flux/sdev"][:]

# Neutron density
n = np.zeros(K)
n_sd = np.zeros(K)
for k in range(K):
    n[k] = np.sum(phi[:, k] / speed) / dt[k]
    n_sd[k] = np.linalg.norm(phi_sd[:, k] / speed) / dt[k]

# Normalize
for k in range(K):
    phi_ref[:, k] *= E_mid / dE
    phi[:, k] *= E_mid / dE / dt[k]
    phi_sd[:, k] *= E_mid / dE / dt[k]

# Flux - t
fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(8, 4))
fig.tight_layout(pad=3.0)

ax1.plot(t_mid, n, "-b", label="MC")
no = np.zeros_like(n)
(line,) = ax1.plot(t[1:], no, "ko", fillstyle="none")
ax1.fill_between(t[1:], n - n_sd, n + n_sd, alpha=0.2, color="b")
ax1.plot(t_mid, n_ref, "--r", label="Ref.")
ax1.set_xlabel(r"$t$, s")
ax1.set_ylabel("Density")
ax1.set_yscale("log")
ax1.set_xscale("log")
ax1.grid()
ax1.legend()
ax1.set_title(r"$n_g(t)$")

ax2.grid()
ax2.set_xlabel(r"$E$, MeV")
ax2.set_ylabel(r"$E\phi(E)$")
ax2.set_title(r"$\phi(E,t)$")
ax2.set_xscale("log")
(line1,) = ax2.plot([], [], "-b", label="MC")
(line2,) = ax2.plot([], [], "--r", label="Ref.")
fb = ax2.fill_between([], [], [], [], alpha=0.2, color="b")
ax2.legend()


def animate(k):
    global fb
    fb.remove()
    no[k - 1] = 0.0
    no[k] = n[k]
    line.set_data(t[1:], no)
    line1.set_data(E_mid, phi[:, k])
    fb = ax2.fill_between(
        E_mid,
        phi[:, k] - phi_sd[:, k],
        phi[:, k] + phi_sd[:, k],
        alpha=0.2,
        color="b",
    )
    line2.set_data(E_mid, phi_ref[:, k])
    ax2.set_ylim([(phi_ref[:, k]).min(), (phi_ref[:, k]).max()])
    ax2.legend()
    return line1, line2


simulation = animation.FuncAnimation(fig, animate, frames=K)
writervideo = animation.FFMpegWriter(fps=6)
plt.show()



import matplotlib.pyplot as plt
import h5py
import numpy as np

# Load iqmc result
with h5py.File("output.h5", "r") as f:
    meshx = f["iqmc/grid/x"][:]
    meshy = f["iqmc/grid/y"][:]
    dx = [meshx[1:] - meshx[:-1]][-1]
    x_mid = 0.5 * (meshx[:-1] + meshx[1:])
    phi = f["iqmc/tally/flux"][:]

    f.close()

# =============================================================================
# Flux Plot
# =============================================================================
X, Y = np.meshgrid(x_mid, x_mid)
Z = np.log10(np.abs(phi / phi.min()))
fig, ax = plt.subplots(subplot_kw={"projection": "3d"}, dpi=300, figsize=(12, 10))
ax.plot_surface(Y, X, Z, edgecolor="b", color="white", linewidth=0.5)
ax.set_xlabel("x")
ax.set_ylabel("y")
ax.set_zlabel(r"log($\phi$)", rotation=180)

ax.view_init(elev=15, azim=20)
plt.show()



import numpy as np
import os
import sys

os.system("rm results/*png")

# Fixed source
for task in os.scandir("./fixed_source"):
    os.chdir(task)
    os.system("rm output*")
    os.chdir(r"../..")

# Eigenvalue
"""
for task in os.scandir('./eigenvalue'):
    os.chdir(task)
    os.system("rm output*")
    os.chdir(r"../..")
"""



import numpy as np
import matplotlib.pyplot as plt
import h5py, sys
from matplotlib import cm
from matplotlib import colors


# =============================================================================
# Plot results
# =============================================================================

# MPACT solution
data = np.loadtxt("mpact_TD4-4.txt")
t_ref = data[:, 0]
phi_ref = data[:, 1]

# Grids
with h5py.File("output_td.h5", "r") as f:
    t = f["tally/grid/t"][:]

# Solution
with h5py.File("output_ic.h5", "r") as f:
    norm = f["IC/fission"][()]
with h5py.File("output_td.h5", "r") as f:
    phi = f["tally/fission/mean"][:]
    phi = phi / norm

plt.figure(figsize=(4, 3))
# plt.plot(t_ref, phi_ref,'r-', label='MPACT')

x = 0.5 * (t[1:] + t[:-1])
y = phi
plt.plot(x, y, "bo", fillstyle="none", label="MC/DC")
plt.xlim([0, 8])
plt.xlabel("Time [s]")
plt.ylabel("Normalized fission rate")
plt.grid()
plt.legend()
plt.savefig("C5G7-TD4-4.png", dpi=1200, bbox_inches="tight", pad_inches=0)
plt.show()



import mcdc
import mcdc.type_
import numpy as np


# ======================================================================================
# Nuclide
# ======================================================================================


def test_nuclide_basic():
    # Start fresh
    mcdc.reset_cards()

    # Create a nuclide with complete definition
    n1 = mcdc.nuclide(
        capture=np.array([1.0, 2.0, 3.0, 4.0]),
        fission=np.array([1.0, 2.0, 3.0, 4.0]),
        scatter=np.array(
            [
                [1.0, 2.0, 3.0, 4.0],
                [1.0, 2.0, 3.0, 4.0],
                [1.0, 2.0, 3.0, 4.0],
                [1.0, 2.0, 3.0, 4.0],
            ]
        ),
        nu_s=np.array([1.0, 2.0, 3.0, 4.0]),
        nu_p=np.array([1.0, 2.0, 3.0, 4.0]),
        nu_d=np.array(
            [
                [1.0, 2.0, 3.0, 4.0],
                [1.0, 2.0, 3.0, 4.0],
                [1.0, 2.0, 3.0, 4.0],
            ]
        ),
        chi_p=np.array(
            [
                [1.0, 2.0, 3.0, 4.0],
                [1.0, 2.0, 3.0, 4.0],
                [1.0, 2.0, 3.0, 4.0],
                [1.0, 2.0, 3.0, 4.0],
            ]
        ),
        chi_d=np.array(
            [
                [1.0, 2.0, 3.0],
                [1.0, 2.0, 3.0],
                [1.0, 2.0, 3.0],
                [1.0, 2.0, 3.0],
            ]
        ),
        speed=np.array([1.0, 2.0, 3.0, 4.0]),
        decay=np.array([1.0, 2.0, 3.0]),
        sensitivity=True,
    )

    # Checks
    assert n1["tag"] == "Nuclide"
    assert n1["ID"] == 0
    assert n1["G"] == 4
    assert n1["J"] == 3
    assert (n1["speed"] == np.array([1.0, 2.0, 3.0, 4.0])).all()
    assert (n1["decay"] == np.array([1.0, 2.0, 3.0])).all()
    assert (n1["capture"] == np.array([1.0, 2.0, 3.0, 4.0])).all()
    assert (n1["scatter"] == np.array([4.0, 8.0, 12.0, 16.0])).all()
    assert (n1["fission"] == np.array([1.0, 2.0, 3.0, 4.0])).all()
    assert (n1["total"] == np.array([6.0, 12.0, 18.0, 24.0])).all()
    assert (n1["nu_s"] == np.array([1.0, 2.0, 3.0, 4.0])).all()
    assert (n1["nu_p"] == np.array([1.0, 2.0, 3.0, 4.0])).all()
    assert (
        n1["nu_d"]
        == np.array(
            [
                [1.0, 1.0, 1.0],
                [2.0, 2.0, 2.0],
                [3.0, 3.0, 3.0],
                [4.0, 4.0, 4.0],
            ]
        )
    ).all()
    assert (n1["nu_f"] == np.array([4.0, 8.0, 12.0, 16.0])).all()
    assert (n1["chi_s"] == np.ones([4, 4]) * 0.25).all()
    assert (n1["chi_p"] == np.ones([4, 4]) * 0.25).all()
    assert (n1["chi_d"] == np.ones([3, 4]) * 0.25).all()
    assert n1["sensitivity"] == True
    assert n1["sensitivity_ID"] == 1


def test_nuclide_default():
    # Start fresh
    mcdc.reset_cards()

    # Create nuclides with incomplete definitions
    n1 = mcdc.nuclide(capture=np.ones(5))
    n2 = mcdc.nuclide(scatter=np.ones((5, 5)))
    n3 = mcdc.nuclide(fission=np.ones(5), nu_p=np.ones(5), chi_p=np.ones((5, 5)))

    # Checks
    assert n1["tag"] == "Nuclide"
    assert n1["ID"] == 0
    assert n1["G"] == 5
    assert n1["J"] == 0
    assert (n1["speed"] == np.ones(5)).all()
    assert (n1["decay"] == np.zeros(0)).all()
    assert (n1["capture"] == np.ones(5)).all()
    assert (n1["scatter"] == np.zeros(5)).all()
    assert (n1["fission"] == np.zeros(5)).all()
    assert (n1["nu_s"] == np.ones(5)).all()
    assert (n1["nu_p"] == np.zeros(5)).all()
    assert (n1["nu_f"] == np.zeros(5)).all()
    assert (n1["nu_d"] == np.zeros((5, 0))).all()
    assert (n1["chi_s"] == np.zeros((5, 5))).all()
    assert (n1["chi_p"] == np.zeros((5, 5))).all()
    assert (n1["chi_d"] == np.zeros((0, 5))).all()
    assert n1["sensitivity"] == False
    assert n1["sensitivity_ID"] == 0

    assert (n2["capture"] == np.zeros(5)).all()
    assert (n2["scatter"] == np.ones(5) * 5.0).all()
    assert (n2["fission"] == np.zeros(5)).all()
    assert (n1["nu_s"] == np.ones(5)).all()
    assert (n2["nu_p"] == np.zeros(5)).all()
    assert (n2["nu_f"] == np.zeros(5)).all()
    assert (n2["nu_d"] == np.zeros((5, 0))).all()
    assert (n2["chi_s"] == np.ones((5, 5)) * 0.2).all()
    assert (n2["chi_p"] == np.zeros((5, 5))).all()
    assert (n2["chi_d"] == np.zeros((0, 5))).all()
    assert n2["sensitivity"] == False
    assert n2["sensitivity_ID"] == 0

    assert (n3["capture"] == np.zeros(5)).all()
    assert (n3["scatter"] == np.zeros(5)).all()
    assert (n3["fission"] == np.ones(5)).all()
    assert (n1["nu_s"] == np.ones(5)).all()
    assert (n3["nu_p"] == np.ones(5)).all()
    assert (n3["nu_f"] == np.ones(5)).all()
    assert (n3["nu_d"] == np.zeros((5, 0))).all()
    assert (n3["chi_s"] == np.zeros((5, 5))).all()
    assert (n3["chi_p"] == np.ones((5, 5)) * 0.2).all()
    assert (n3["chi_d"] == np.zeros((0, 5))).all()
    assert n3["sensitivity"] == False
    assert n3["sensitivity_ID"] == 0


def test_nuclide_IDs_sensitivity():
    # Start fresh
    mcdc.reset_cards()

    # Create nuclides with various sensitivity tags
    n1 = mcdc.nuclide(capture=np.array([1.0]), sensitivity=True)
    n2 = mcdc.nuclide(capture=np.array([1.0]))
    n3 = mcdc.nuclide(capture=np.array([1.0]), sensitivity=True)
    n4 = mcdc.nuclide(capture=np.array([1.0]), sensitivity=True)
    n5 = mcdc.nuclide(capture=np.array([1.0]))
    n6 = mcdc.nuclide(capture=np.array([1.0]))
    n7 = mcdc.nuclide(capture=np.array([1.0]), sensitivity=True)

    # Checks
    assert n1["ID"] == 0
    assert n2["ID"] == 1
    assert n3["ID"] == 2
    assert n4["ID"] == 3
    assert n5["ID"] == 4
    assert n6["ID"] == 5
    assert n7["ID"] == 6
    assert n1["sensitivity"]
    assert not n2["sensitivity"]
    assert n3["sensitivity"]
    assert n4["sensitivity"]
    assert not n5["sensitivity"]
    assert not n6["sensitivity"]
    assert n7["sensitivity"]
    assert n1["sensitivity_ID"] == 1
    assert n2["sensitivity_ID"] == 0
    assert n3["sensitivity_ID"] == 2
    assert n4["sensitivity_ID"] == 3
    assert n5["sensitivity_ID"] == 0
    assert n6["sensitivity_ID"] == 0
    assert n7["sensitivity_ID"] == 4


# ======================================================================================
# Material
# ======================================================================================


def test_material_single():
    # Start fresh
    mcdc.reset_cards()

    # Create a single-nuclide material
    m1 = mcdc.material(
        capture=np.array([1.0, 2.0, 3.0, 4.0]),
        fission=np.array([1.0, 2.0, 3.0, 4.0]),
        scatter=np.array(
            [
                [1.0, 2.0, 3.0, 4.0],
                [1.0, 2.0, 3.0, 4.0],
                [1.0, 2.0, 3.0, 4.0],
                [1.0, 2.0, 3.0, 4.0],
            ]
        ),
        nu_s=np.array([1.0, 2.0, 3.0, 4.0]),
        nu_p=np.array([1.0, 2.0, 3.0, 4.0]),
        nu_d=np.array(
            [
                [1.0, 2.0, 3.0, 4.0],
                [1.0, 2.0, 3.0, 4.0],
                [1.0, 2.0, 3.0, 4.0],
            ]
        ),
        chi_p=np.array(
            [
                [1.0, 2.0, 3.0, 4.0],
                [1.0, 2.0, 3.0, 4.0],
                [1.0, 2.0, 3.0, 4.0],
                [1.0, 2.0, 3.0, 4.0],
            ]
        ),
        chi_d=np.array(
            [
                [1.0, 2.0, 3.0],
                [1.0, 2.0, 3.0],
                [1.0, 2.0, 3.0],
                [1.0, 2.0, 3.0],
            ]
        ),
        speed=np.array([1.0, 2.0, 3.0, 4.0]),
        decay=np.array([1.0, 2.0, 3.0]),
        sensitivity=True,
    )

    # Checks
    assert m1["tag"] == "Material"
    assert m1["ID"] == 0
    assert m1["N_nuclide"] == 1
    assert (m1["nuclide_IDs"] == np.array([0])).all()
    assert (m1["nuclide_densities"] == np.array([1.0])).all()
    assert m1["G"] == 4
    assert m1["J"] == 3
    assert (m1["speed"] == np.array([1.0, 2.0, 3.0, 4.0])).all()
    assert (m1["capture"] == np.array([1.0, 2.0, 3.0, 4.0])).all()
    assert (m1["scatter"] == np.array([4.0, 8.0, 12.0, 16.0])).all()
    assert (m1["fission"] == np.array([1.0, 2.0, 3.0, 4.0])).all()
    assert (m1["total"] == np.array([6.0, 12.0, 18.0, 24.0])).all()
    assert (m1["nu_s"] == np.array([1.0, 2.0, 3.0, 4.0])).all()
    assert (m1["nu_p"] == np.array([1.0, 2.0, 3.0, 4.0])).all()
    assert (
        m1["nu_d"]
        == np.array(
            [
                [1.0, 1.0, 1.0],
                [2.0, 2.0, 2.0],
                [3.0, 3.0, 3.0],
                [4.0, 4.0, 4.0],
            ]
        )
    ).all()
    assert (m1["nu_f"] == np.array([4.0, 8.0, 12.0, 16.0])).all()
    assert (m1["chi_s"] == np.ones([4, 4]) * 0.25).all()
    assert (m1["chi_p"] == np.ones([4, 4]) * 0.25).all()
    assert m1["sensitivity"] == True

    # Check if the nuclide was registered
    n2 = mcdc.nuclide(capture=np.ones(5))
    assert n2["ID"] == 1


def test_material_multi():
    # Start fresh
    mcdc.reset_cards()

    # Create a multi-nuclide material
    n1 = mcdc.nuclide(capture=np.ones(5), speed=np.ones(5) * 1)
    n2 = mcdc.nuclide(scatter=np.ones((5, 5)), speed=np.ones(5) * 2, sensitivity=True)
    n3 = mcdc.nuclide(
        fission=np.ones(5), nu_p=np.ones(5), chi_p=np.ones((5, 5)), speed=np.ones(5) * 3
    )
    m1 = mcdc.material(nuclides=[(n1, 1.0), (n2, 2.0), (n3, 3.0)])

    # Checks
    assert m1["tag"] == "Material"
    assert m1["ID"] == 0
    assert m1["N_nuclide"] == 3
    assert (m1["nuclide_IDs"] == np.array([0, 1, 2])).all()
    assert (m1["nuclide_densities"] == np.array([1.0, 2.0, 3.0])).all()
    assert m1["G"] == 5
    assert m1["J"] == 0
    assert (m1["total"] == np.ones(5) * 14).all()
    assert (m1["speed"] == np.ones(5) * 30 / 14).all()
    assert m1["sensitivity"] == True


def test_material_IDs_sensitivity():
    # Start fresh
    mcdc.reset_cards()

    # Create materials with various sensitivity tags
    m1 = mcdc.material(capture=np.array([1.0]), sensitivity=True)
    m2 = mcdc.material(capture=np.array([1.0]))
    m3 = mcdc.material(capture=np.array([1.0]), sensitivity=True)
    m4 = mcdc.material(capture=np.array([1.0]), sensitivity=True)
    m5 = mcdc.material(capture=np.array([1.0]))
    m6 = mcdc.material(capture=np.array([1.0]))
    m7 = mcdc.material(capture=np.array([1.0]), sensitivity=True)

    # Checks
    assert m1["ID"] == 0
    assert m2["ID"] == 1
    assert m3["ID"] == 2
    assert m4["ID"] == 3
    assert m5["ID"] == 4
    assert m6["ID"] == 5
    assert m7["ID"] == 6
    assert m1["sensitivity"]
    assert not m2["sensitivity"]
    assert m3["sensitivity"]
    assert m4["sensitivity"]
    assert not m5["sensitivity"]
    assert not m6["sensitivity"]
    assert m7["sensitivity"]


# ======================================================================================
# Surface
# ======================================================================================


def test_surface_input_lower():
    type_ = "pLaNe x"
    result = mcdc.surface(type_, bc="RefLeCtiVe", x=0.0)
    assert result.card["reflective"]
    assert result.card["A"] == 0.0
    assert result.card["B"] == 0.0
    assert result.card["C"] == 0.0
    assert result.card["D"] == 0.0
    assert result.card["E"] == 0.0
    assert result.card["F"] == 0.0
    assert result.card["G"] == 1.0
    assert result.card["H"] == 0.0
    assert result.card["I"] == 0.0
    assert (result.card["J"] == np.array([[-0.0, 0.0]])).all()
    assert result.card["linear"]


# ======================================================================================
# Reset
# ======================================================================================


def test_reset_cards():
    # Start fresh
    mcdc.reset_cards()

    # ID reset
    n = mcdc.nuclide(capture=np.ones(5))
    assert n["ID"] == 0

    mcdc.reset_cards()

    n1 = mcdc.nuclide(capture=np.ones(5))
    assert n1["ID"] == 0



import numpy as np
import matplotlib.pyplot as plt
import h5py


# =============================================================================
# Reference solution
# =============================================================================

# Load XS
with np.load("SHEM-361.npz") as data:
    data = np.load("SHEM-361.npz")
    SigmaT = data["SigmaT"]
    nuSigmaF = data["nuSigmaF"]
    SigmaS = data["SigmaS"]
    E = data["E"]

G = len(SigmaT)
E_mid = 0.5 * (E[1:] + E[:-1])
dE = E[1:] - E[:-1]

A = np.dot(np.linalg.inv(np.diag(SigmaT) - SigmaS), nuSigmaF)
w, v = np.linalg.eig(A)
idx = w.argsort()[::-1]
w = w[idx]
v = v[:, idx]
k_exact = w[0]
phi_exact = v[:, 0]
if phi_exact[0] < 0.0:
    phi_exact *= -1
    phi_exact[:] /= np.sum(phi_exact[:])
    phi_exact = phi_exact / dE * E_mid

# =============================================================================
# Plot results
# =============================================================================

# Results
with h5py.File("output.h5", "r") as f:
    phi_avg = f["tally/flux/mean"][:]
    phi_sd = f["tally/flux/sdev"][:]
    k = f["k_cycle"][:]
    k_avg = f["k_mean"][()]
    k_sd = f["k_sdev"][()]

norm = np.sum(phi_avg)
phi_avg = phi_avg / norm / dE * E_mid
phi_sd = phi_sd / norm / dE * E_mid

# Plot
N_iter = len(k)
plt.plot(np.arange(2, N_iter + 1), k[1:], "-b", label="MC")
plt.plot(
    np.arange(2, N_iter + 1), np.ones(N_iter - 1) * k_exact, "--r", label="analytical"
)
plt.plot(np.arange(2, N_iter + 1), np.ones(N_iter - 1) * k_avg, ":g", label="MC-avg")
plt.fill_between(
    np.arange(2, N_iter + 1),
    np.ones(N_iter - 1) * (k_avg - k_sd),
    np.ones(N_iter - 1) * (k_avg + k_sd),
    alpha=0.2,
    color="g",
)
plt.xlabel("Iteration #")
plt.ylabel(r"$k$")
plt.grid()
plt.legend()
plt.show()

plt.step(E_mid, phi_avg, "-b", label="MC", where="mid")
plt.fill_between(
    E_mid, phi_avg - phi_sd, phi_avg + phi_sd, alpha=0.2, color="b", step="mid"
)
plt.step(E_mid, phi_exact, "--r", label="analytical", where="mid")
plt.xscale("log")
plt.xlabel(r"$E$, eV")
plt.ylabel(r"$E\phi(E)$")
plt.grid()
plt.legend()
plt.show()



import h5py
import numpy as np

import mcdc

# =============================================================================
# Materials
# =============================================================================

# Load material data
lib = h5py.File("c5g7.h5", "r")


# Materials
def set_mat(mat):
    return mcdc.material(
        capture=mat["capture"][:],
        scatter=mat["scatter"][:],
        fission=mat["fission"][:],
        nu_p=mat["nu_p"][:],
        nu_d=mat["nu_d"][:],
        chi_p=mat["chi_p"][:],
        chi_d=mat["chi_d"][:],
        speed=mat["speed"],
        decay=mat["decay"],
    )


mat_uo2 = set_mat(lib["uo2"])
mat_mox43 = set_mat(lib["mox43"])
mat_mox7 = set_mat(lib["mox7"])
mat_mox87 = set_mat(lib["mox87"])
mat_gt = set_mat(lib["gt"])
mat_fc = set_mat(lib["fc"])
mat_cr = set_mat(lib["cr"])
mat_mod = set_mat(lib["mod"])

# =============================================================================
# Pin cells
# =============================================================================

pitch = 1.26
radius = 0.54

# Surfaces
cy = mcdc.surface("cylinder-z", center=[0.0, 0.0], radius=radius)
z0 = mcdc.surface("plane-z", z=-100.0)
z1 = mcdc.surface("plane-z", z=100.0)

# Cells
fc = mcdc.cell([-cy, +z0], mat_fc)  # Fission chamber (in and above core)
uo2 = mcdc.cell([-cy, +z0, -z1], mat_uo2)  # Fuel rods (in core)
mox4 = mcdc.cell([-cy, +z0, -z1], mat_mox43)
mox7 = mcdc.cell([-cy, +z0, -z1], mat_mox7)
mox8 = mcdc.cell([-cy, +z0, -z1], mat_mox87)
cr = mcdc.cell([-cy, +z1], mat_cr)  # Control rod (above core)
gti = mcdc.cell([-cy, +z0, -z1], mat_gt)  # Guide tube (in core)
gta = mcdc.cell([-cy, +z1], mat_gt)  #            (above core)
mod = mcdc.cell([+cy], mat_mod)  # Moderator (outside rod)
modu = mcdc.cell([-cy, -z0], mat_mod)  #           (under rod)

# Universes (pin cells)
f = mcdc.universe([fc, mod, modu])["ID"]
u = mcdc.universe([uo2, gta, mod, modu])["ID"]
l = mcdc.universe([mox4, gta, mod, modu])["ID"]
m = mcdc.universe([mox7, gta, mod, modu])["ID"]
n = mcdc.universe([mox8, gta, mod, modu])["ID"]
c = mcdc.universe([gti, cr, mod, modu])["ID"]
g = mcdc.universe([gti, gta, mod, modu])["ID"]

# =============================================================================
# Assemblies
# =============================================================================

# Lattices
lattice_uo2 = mcdc.lattice(
    x=[-pitch * 17 / 2, pitch, 17],
    y=[-pitch * 17 / 2, pitch, 17],
    universes=[
        [u, u, u, u, u, u, u, u, u, u, u, u, u, u, u, u, u],
        [u, u, u, u, u, u, u, u, u, u, u, u, u, u, u, u, u],
        [u, u, u, u, u, g, u, u, g, u, u, g, u, u, u, u, u],
        [u, u, u, g, u, u, u, u, u, u, u, u, u, g, u, u, u],
        [u, u, u, u, u, u, u, u, u, u, u, u, u, u, u, u, u],
        [u, u, g, u, u, g, u, u, g, u, u, g, u, u, g, u, u],
        [u, u, u, u, u, u, u, u, u, u, u, u, u, u, u, u, u],
        [u, u, u, u, u, u, u, u, u, u, u, u, u, u, u, u, u],
        [u, u, g, u, u, g, u, u, f, u, u, g, u, u, g, u, u],
        [u, u, u, u, u, u, u, u, u, u, u, u, u, u, u, u, u],
        [u, u, u, u, u, u, u, u, u, u, u, u, u, u, u, u, u],
        [u, u, g, u, u, g, u, u, g, u, u, g, u, u, g, u, u],
        [u, u, u, u, u, u, u, u, u, u, u, u, u, u, u, u, u],
        [u, u, u, g, u, u, u, u, u, u, u, u, u, g, u, u, u],
        [u, u, u, u, u, g, u, u, g, u, u, g, u, u, u, u, u],
        [u, u, u, u, u, u, u, u, u, u, u, u, u, u, u, u, u],
        [u, u, u, u, u, u, u, u, u, u, u, u, u, u, u, u, u],
    ],
)

lattice_mox = mcdc.lattice(
    x=[-pitch * 17 / 2, pitch, 17],
    y=[-pitch * 17 / 2, pitch, 17],
    universes=[
        [l, l, l, l, l, l, l, l, l, l, l, l, l, l, l, l, l],
        [l, m, m, m, m, m, m, m, m, m, m, m, m, m, m, m, l],
        [l, m, m, m, m, g, m, m, g, m, m, g, m, m, m, m, l],
        [l, m, m, g, m, n, n, n, n, n, n, n, m, g, m, m, l],
        [l, m, m, m, n, n, n, n, n, n, n, n, n, m, m, m, l],
        [l, m, g, n, n, g, n, n, g, n, n, g, n, n, g, m, l],
        [l, m, m, n, n, n, n, n, n, n, n, n, n, n, m, m, l],
        [l, m, m, n, n, n, n, n, n, n, n, n, n, n, m, m, l],
        [l, m, g, n, n, g, n, n, f, n, n, g, n, n, g, m, l],
        [l, m, m, n, n, n, n, n, n, n, n, n, n, n, m, m, l],
        [l, m, m, n, n, n, n, n, n, n, n, n, n, n, m, m, l],
        [l, m, g, n, n, g, n, n, g, n, n, g, n, n, g, m, l],
        [l, m, m, m, n, n, n, n, n, n, n, n, n, m, m, m, l],
        [l, m, m, g, m, n, n, n, n, n, n, n, m, g, m, m, l],
        [l, m, m, m, m, g, m, m, g, m, m, g, m, m, m, m, l],
        [l, m, m, m, m, m, m, m, m, m, m, m, m, m, m, m, l],
        [l, l, l, l, l, l, l, l, l, l, l, l, l, l, l, l, l],
    ],
)

lattice_uo2_cr = mcdc.lattice(
    x=[-pitch * 17 / 2, pitch, 17],
    y=[-pitch * 17 / 2, pitch, 17],
    universes=[
        [u, u, u, u, u, u, u, u, u, u, u, u, u, u, u, u, u],
        [u, u, u, u, u, u, u, u, u, u, u, u, u, u, u, u, u],
        [u, u, u, u, u, c, u, u, c, u, u, c, u, u, u, u, u],
        [u, u, u, c, u, u, u, u, u, u, u, u, u, c, u, u, u],
        [u, u, u, u, u, u, u, u, u, u, u, u, u, u, u, u, u],
        [u, u, c, u, u, c, u, u, c, u, u, c, u, u, c, u, u],
        [u, u, u, u, u, u, u, u, u, u, u, u, u, u, u, u, u],
        [u, u, u, u, u, u, u, u, u, u, u, u, u, u, u, u, u],
        [u, u, c, u, u, c, u, u, f, u, u, c, u, u, c, u, u],
        [u, u, u, u, u, u, u, u, u, u, u, u, u, u, u, u, u],
        [u, u, u, u, u, u, u, u, u, u, u, u, u, u, u, u, u],
        [u, u, c, u, u, c, u, u, c, u, u, c, u, u, c, u, u],
        [u, u, u, u, u, u, u, u, u, u, u, u, u, u, u, u, u],
        [u, u, u, c, u, u, u, u, u, u, u, u, u, c, u, u, u],
        [u, u, u, u, u, c, u, u, c, u, u, c, u, u, u, u, u],
        [u, u, u, u, u, u, u, u, u, u, u, u, u, u, u, u, u],
        [u, u, u, u, u, u, u, u, u, u, u, u, u, u, u, u, u],
    ],
)

lattice_mox_cr = mcdc.lattice(
    x=[-pitch * 17 / 2, pitch, 17],
    y=[-pitch * 17 / 2, pitch, 17],
    universes=[
        [l, l, l, l, l, l, l, l, l, l, l, l, l, l, l, l, l],
        [l, m, m, m, m, m, m, m, m, m, m, m, m, m, m, m, l],
        [l, m, m, m, m, c, m, m, c, m, m, c, m, m, m, m, l],
        [l, m, m, c, m, n, n, n, n, n, n, n, m, c, m, m, l],
        [l, m, m, m, n, n, n, n, n, n, n, n, n, m, m, m, l],
        [l, m, c, n, n, c, n, n, c, n, n, c, n, n, c, m, l],
        [l, m, m, n, n, n, n, n, n, n, n, n, n, n, m, m, l],
        [l, m, m, n, n, n, n, n, n, n, n, n, n, n, m, m, l],
        [l, m, c, n, n, c, n, n, f, n, n, c, n, n, c, m, l],
        [l, m, m, n, n, n, n, n, n, n, n, n, n, n, m, m, l],
        [l, m, m, n, n, n, n, n, n, n, n, n, n, n, m, m, l],
        [l, m, c, n, n, c, n, n, c, n, n, c, n, n, c, m, l],
        [l, m, m, m, n, n, n, n, n, n, n, n, n, m, m, m, l],
        [l, m, m, c, m, n, n, n, n, n, n, n, m, c, m, m, l],
        [l, m, m, m, m, c, m, m, c, m, m, c, m, m, m, m, l],
        [l, m, m, m, m, m, m, m, m, m, m, m, m, m, m, m, l],
        [l, l, l, l, l, l, l, l, l, l, l, l, l, l, l, l, l],
    ],
)

# Reflector
gt = mcdc.cell([-cy], mat_gt)  # Reflector
gto = mcdc.cell([+cy], mat_gt)  #
modi = mcdc.cell([-cy], mat_mod)  # Moderator (inside rod)
g = mcdc.universe([gt, gto])["ID"]
w = mcdc.universe([modi, mod])["ID"]
lattice_r = mcdc.lattice(
    x=[-pitch * 17 / 2, pitch, 17],
    y=[-pitch * 17 / 2, pitch, 17],
    universes=[
        [g, g, g, g, g, g, g, g, g, g, g, g, g, g, g, g, g],
        [g, g, g, g, g, g, g, g, g, g, g, g, g, g, g, g, g],
        [w, g, g, w, g, g, w, g, g, w, g, g, w, g, g, w, g],
        [g, g, g, g, g, g, g, g, g, g, g, g, g, g, g, g, g],
        [g, g, g, g, g, g, g, g, g, g, g, g, g, g, g, g, g],
        [g, w, g, g, w, g, g, w, g, g, w, g, g, w, g, g, w],
        [g, g, g, g, g, g, g, g, g, g, g, g, g, g, g, g, g],
        [g, g, g, g, g, g, g, g, g, g, g, g, g, g, g, g, g],
        [g, g, w, g, g, w, g, g, w, g, g, w, g, g, w, g, g],
        [g, g, g, g, g, g, g, g, g, g, g, g, g, g, g, g, g],
        [g, g, g, g, g, g, g, g, g, g, g, g, g, g, g, g, g],
        [w, g, g, w, g, g, w, g, g, w, g, g, w, g, g, w, g],
        [g, g, g, g, g, g, g, g, g, g, g, g, g, g, g, g, g],
        [g, g, g, g, g, g, g, g, g, g, g, g, g, g, g, g, g],
        [g, w, g, g, w, g, g, w, g, g, w, g, g, w, g, g, w],
        [g, g, g, g, g, g, g, g, g, g, g, g, g, g, g, g, g],
        [g, g, g, g, g, g, g, g, g, g, g, g, g, g, g, g, g],
    ],
)

lattice_rll = mcdc.lattice(
    x=[-pitch * 17 / 2, pitch, 17],
    y=[-pitch * 17 / 2, pitch, 17],
    universes=[
        [g, g, g, g, g, g, g, g, g, g, g, g, g, g, g, g, g],
        [g, g, g, g, g, g, g, g, g, g, g, g, g, g, g, g, g],
        [g, g, g, g, g, g, g, g, g, g, g, g, g, g, g, g, g],
        [w, g, g, g, w, g, g, g, w, g, g, w, w, w, g, g, g],
        [g, g, g, g, g, g, g, g, g, g, g, w, w, w, g, g, g],
        [g, g, g, g, g, g, g, g, g, g, g, w, w, w, g, g, g],
        [g, g, g, g, g, g, g, g, g, g, g, g, g, g, g, g, g],
        [g, g, g, g, g, g, g, g, g, w, g, g, g, g, g, g, g],
        [g, g, g, g, g, g, w, g, g, g, g, g, g, w, g, g, g],
        [g, g, g, w, g, g, g, g, g, g, g, g, g, g, g, g, g],
        [g, g, g, g, g, g, g, g, w, g, g, g, g, g, g, g, g],
        [g, g, g, g, g, g, g, g, g, g, g, g, g, g, g, g, g],
        [g, g, g, g, w, g, g, g, g, g, g, g, g, w, g, g, g],
        [g, g, g, g, g, g, g, w, g, g, g, g, g, g, g, g, g],
        [g, g, g, g, g, g, g, g, g, g, g, g, g, g, g, g, g],
        [g, g, g, g, g, g, g, g, g, g, g, g, g, g, g, g, g],
        [w, g, g, g, g, g, g, g, g, g, g, g, g, w, g, g, g],
    ],
)

lattice_rul = mcdc.lattice(
    x=[-pitch * 17 / 2, pitch, 17],
    y=[-pitch * 17 / 2, pitch, 17],
    universes=[
        [w, g, g, g, g, g, g, g, g, g, g, g, g, w, g, g, g],
        [g, g, g, g, g, g, g, g, g, g, g, g, g, g, g, g, g],
        [g, g, g, g, g, g, g, g, g, g, g, g, g, g, g, g, g],
        [g, g, g, g, g, g, g, w, g, g, g, g, g, g, g, g, g],
        [g, g, g, g, w, g, g, g, g, g, g, g, g, w, g, g, g],
        [g, g, g, g, g, g, g, g, g, g, g, g, g, g, g, g, g],
        [g, g, g, g, g, g, g, g, w, g, g, g, g, g, g, g, g],
        [g, g, g, w, g, g, g, g, g, g, g, g, g, g, g, g, g],
        [g, g, g, g, g, g, w, g, g, g, g, g, g, w, g, g, g],
        [g, g, g, g, g, g, g, g, g, w, g, g, g, g, g, g, g],
        [g, g, g, g, g, g, g, g, g, g, g, g, g, g, g, g, g],
        [g, g, g, g, g, g, g, g, g, g, g, w, w, w, g, g, g],
        [g, g, g, g, g, g, g, g, g, g, g, w, w, w, g, g, g],
        [w, g, g, g, w, g, g, g, w, g, g, w, w, w, g, g, g],
        [g, g, g, g, g, g, g, g, g, g, g, g, g, g, g, g, g],
        [g, g, g, g, g, g, g, g, g, g, g, g, g, g, g, g, g],
        [g, g, g, g, g, g, g, g, g, g, g, g, g, g, g, g, g],
    ],
)

lattice_rur = mcdc.lattice(
    x=[-pitch * 17 / 2, pitch, 17],
    y=[-pitch * 17 / 2, pitch, 17],
    universes=[
        [g, g, g, w, g, g, g, g, g, g, g, g, g, g, g, g, w],
        [g, g, g, g, g, g, g, g, g, g, g, g, g, g, g, g, g],
        [g, g, g, g, g, g, g, g, g, g, g, g, g, g, g, g, g],
        [g, g, g, g, g, g, g, g, g, w, g, g, g, g, g, g, g],
        [g, g, g, w, g, g, g, g, g, g, g, g, w, g, g, g, g],
        [g, g, g, g, g, g, g, g, g, g, g, g, g, g, g, g, g],
        [g, g, g, g, g, g, g, g, w, g, g, g, g, g, g, g, g],
        [g, g, g, g, g, g, g, g, g, g, g, g, g, w, g, g, g],
        [g, g, g, w, g, g, g, g, g, g, w, g, g, g, g, g, g],
        [g, g, g, g, g, g, g, w, g, g, g, g, g, g, g, g, g],
        [g, g, g, g, g, g, g, g, g, g, g, g, g, g, g, g, g],
        [g, g, g, w, w, w, g, g, g, g, g, g, g, g, g, g, g],
        [g, g, g, w, w, w, g, g, g, g, g, g, g, g, g, g, g],
        [g, g, g, w, w, w, g, g, w, g, g, g, w, g, g, g, w],
        [g, g, g, g, g, g, g, g, g, g, g, g, g, g, g, g, g],
        [g, g, g, g, g, g, g, g, g, g, g, g, g, g, g, g, g],
        [g, g, g, g, g, g, g, g, g, g, g, g, g, g, g, g, g],
    ],
)

lattice_rlr = mcdc.lattice(
    x=[-pitch * 17 / 2, pitch, 17],
    y=[-pitch * 17 / 2, pitch, 17],
    universes=[
        [g, g, g, g, g, g, g, g, g, g, g, g, g, g, g, g, g],
        [g, g, g, g, g, g, g, g, g, g, g, g, g, g, g, g, g],
        [g, g, g, g, g, g, g, g, g, g, g, g, g, g, g, g, g],
        [g, g, g, w, w, w, g, g, w, g, g, g, w, g, g, g, w],
        [g, g, g, w, w, w, g, g, g, g, g, g, g, g, g, g, g],
        [g, g, g, w, w, w, g, g, g, g, g, g, g, g, g, g, g],
        [g, g, g, g, g, g, g, g, g, g, g, g, g, g, g, g, g],
        [g, g, g, g, g, g, g, w, g, g, g, g, g, g, g, g, g],
        [g, g, g, w, g, g, g, g, g, g, w, g, g, g, g, g, g],
        [g, g, g, g, g, g, g, g, g, g, g, g, g, w, g, g, g],
        [g, g, g, g, g, g, g, g, w, g, g, g, g, g, g, g, g],
        [g, g, g, g, g, g, g, g, g, g, g, g, g, g, g, g, g],
        [g, g, g, w, g, g, g, g, g, g, g, g, w, g, g, g, g],
        [g, g, g, g, g, g, g, g, g, w, g, g, g, g, g, g, g],
        [g, g, g, g, g, g, g, g, g, g, g, g, g, g, g, g, g],
        [g, g, g, g, g, g, g, g, g, g, g, g, g, g, g, g, g],
        [g, g, g, w, g, g, g, g, g, g, g, g, g, g, g, g, w],
    ],
)

# Assembly cells
# Surfaces
x0 = mcdc.surface("plane-x", x=-pitch * 17 / 2)
x1 = mcdc.surface("plane-x", x=pitch * 17 / 2)
y0 = mcdc.surface("plane-y", y=-pitch * 17 / 2)
y1 = mcdc.surface("plane-y", y=pitch * 17 / 2)
# Cells
assembly_uo2 = mcdc.cell([+x0, -x1, +y0, -y1], lattice_uo2)
assembly_mox = mcdc.cell([+x0, -x1, +y0, -y1], lattice_mox)
assembly_rll = mcdc.cell([+x0, -x1, +y0, -y1], lattice_rll)
assembly_rul = mcdc.cell([+x0, -x1, +y0, -y1], lattice_rul)
assembly_rur = mcdc.cell([+x0, -x1, +y0, -y1], lattice_rur)
assembly_rlr = mcdc.cell([+x0, -x1, +y0, -y1], lattice_rlr)
assembly_r = mcdc.cell([+x0, -x1, +y0, -y1], lattice_r)

# Set assemblies in their respective universes
u = mcdc.universe([assembly_uo2])["ID"]
m = mcdc.universe([assembly_mox])["ID"]
a = mcdc.universe([assembly_rll])["ID"]
b = mcdc.universe([assembly_rul])["ID"]
c = mcdc.universe([assembly_rur])["ID"]
d = mcdc.universe([assembly_rlr])["ID"]
e = mcdc.universe([assembly_r])["ID"]

# =============================================================================
# Root universe: core
# =============================================================================

# Lattice
lattice_core = mcdc.lattice(
    x=[-pitch * 17 * 9 / 2, pitch * 17, 9],
    y=[-pitch * 17 * 9 / 2, pitch * 17, 9],
    universes=[
        [e, e, e, e, e, e, e, e, e],
        [e, b, b, u, m, u, c, c, e],
        [e, b, m, u, u, u, m, c, e],
        [e, u, u, m, u, m, u, u, e],
        [e, m, u, u, m, u, u, m, e],
        [e, u, u, m, u, m, u, u, e],
        [e, a, m, u, u, u, m, d, e],
        [e, e, a, u, m, u, d, d, e],
        [e, e, e, e, e, e, e, e, e],
    ],
)

# Core cell
# Surfaces
cy0 = mcdc.surface("cylinder-z", center=[0.0, 0.0], radius=pitch * 17 * 8.6 / 2)
cy1 = mcdc.surface("cylinder-z", center=[0.0, 0.0], radius=pitch * 17 * 9 / 2)
cy2 = mcdc.surface("cylinder-z", center=[0.0, 0.0], radius=pitch * 17 * 11 / 2)
cy3 = mcdc.surface(
    "cylinder-z", center=[0.0, 0.0], radius=pitch * 17 * 12 / 2, bc="vacuum"
)
zlo = mcdc.surface("plane-z", z=-130, bc="vacuum")
zup = mcdc.surface("plane-z", z=130, bc="vacuum")
# Cell
core = mcdc.cell([-cy0, +zlo, -zup], lattice_core)
barrel = mcdc.cell([+cy0, -cy1, +zlo, -zup], mat_gt)
water = mcdc.cell([+cy1, -cy2, +zlo, -zup], mat_mod)
vessel = mcdc.cell([+cy2, -cy3, +zlo, -zup], mat_gt)

# Root universe
mcdc.universe([core, barrel, water, vessel], root=True)

# =============================================================================
# Set source
# =============================================================================
# Uniform in energy

source = mcdc.source(
    x=[-pitch * 17 * 7 / 2, pitch * 17 * 7 / 2],
    y=[-pitch * 17 * 7 / 2, pitch * 17 * 7 / 2],
    z=[-100.0, 100.0],
    energy=np.ones(7),
)

# =============================================================================
# Set tally and parameter, and then run mcdc
# =============================================================================

# Tally
x_grid = np.linspace(-pitch * 17 * 9 / 2, pitch * 17 * 9 / 2, 17 * 9 + 1)
y_grid = np.linspace(-pitch * 17 * 9 / 2, pitch * 17 * 9 / 2, 17 * 9 + 1)
z_grid = np.linspace(-130, 130, 131)
mcdc.tally(scores=["flux", "fission"], x=x_grid, y=y_grid, z=z_grid, g="all")

# Setting
mcdc.setting(N_particle=1e2)
mcdc.eigenmode(N_inactive=10, N_active=30, gyration_radius="all")
mcdc.population_control()

# Run
mcdc.run()



import h5py, os, sys, argparse, fnmatch
import numpy as np
from colorama import Fore, Style

# Option parser
parser = argparse.ArgumentParser(description="MC/DC regression test")
parser.add_argument("--mode", type=str, choices=["python", "numba"], default="python")
parser.add_argument("--mpiexec", type=int, default=0)
parser.add_argument("--srun", type=int, default=0)
parser.add_argument("--name", type=str, default="ALL")
parser.add_argument("--skip", type=str, default="NONE")
args, unargs = parser.parse_known_args()

# Parse
mode = args.mode
mpiexec = args.mpiexec
srun = args.srun
name = args.name
skip = args.skip

# Get test names
if name == "ALL":
    names = []
    for item in os.listdir():
        if os.path.isdir(item):
            names.append(item)
else:
    names = [item for item in os.listdir() if fnmatch.fnmatch(item, name)]
names.sort()

# Remove skipped if specified
if skip != "NONE":
    skips = [item for item in os.listdir() if fnmatch.fnmatch(item, skip)]
    for name in skips:
        names.remove(name)

# Data for each test
printouts = []
runtimes = []
flags = []
error_msgs = []
crashes = []
all_pass = True

# Run all tests
for i, name in enumerate(names):
    # Skip cache if any
    if name == "__pycache__":
        continue

    print("\n[%i/%i] " % (i + 1, len(names)) + name)
    error_msgs.append([])
    crashes.append(False)
    runtimes.append(-1)

    # Change directory
    os.chdir(name)

    # Check test setup
    if not os.path.exists("input.py"):
        print(Fore.RED + "  input.py is missing\n" + Style.RESET_ALL)
        sys.exit()
    if not os.path.exists("answer.h5"):
        print(Fore.RED + "  answer.h5 is missing\n" + Style.RESET_ALL)
        sys.exit()

    # Delete output if exists
    if os.path.exists("output.h5"):
        os.remove("output.h5")

    # Run the test problem (redirect the stdout)
    if mpiexec > 1:
        os.system(
            "mpiexec -n %i python input.py --mode=%s --output=output --no-progress-bar > tmp 2>&1"
            % (mpiexec, mode)
        )
    elif srun > 1:
        os.system(
            "srun -n %i python input.py --mode=%s --output=output --no-progress-bar > tmp 2>&1"
            % (srun, mode)
        )
    else:
        os.system(
            "python input.py --mode=%s --output=output --no-progress-bar > tmp 2>&1"
            % (mode)
        )
    with open("tmp") as f:
        printouts.append(f.read())
    os.remove("tmp")

    # Check if crashed
    if not os.path.exists("output.h5"):
        print(Fore.RED + "  Failed: Run crashed" + Style.RESET_ALL)
        all_pass = False
        crashes[-1] = True
        os.chdir("..")
        continue

    # Get the output and the answer key
    output = h5py.File("output.h5", "r")
    answer = h5py.File("answer.h5", "r")

    runtimes[-1] = output["runtime/total"][()]
    print("  (%.2f seconds)" % runtimes[-1])

    # Compare all scores
    for score in [key for key in output["tally"].keys() if key != "grid"]:
        # Compare mean, sdev, and uq_var (if available)
        for result in [key for key in output["tally/" + score].keys()]:
            result_name = "tally/" + score + "/" + result
            a = output[result_name][:]
            b = answer[result_name][:]

            # Passed?
            if np.isclose(a, b).all():
                print(
                    Fore.GREEN
                    + "  {}: Passed".format(score + "/" + result)
                    + Style.RESET_ALL
                )
            else:
                all_pass = False
                error_msgs[-1].append(
                    "Differences in %s"
                    % (name + "/" + score + "/" + result + "\n" + "{}".format(a - b))
                )
                print(
                    Fore.RED
                    + "  {}: Failed".format(score + "/" + result)
                    + Style.RESET_ALL
                )

    # Other quantities
    for result_name in ["k_mean", "k_sdev", "k_cycle", "k_eff"]:
        if result_name not in output.keys():
            continue

        a = output[result_name][()]
        b = answer[result_name][()]

        # Passed?
        if np.isclose(a, b).all():
            print(Fore.GREEN + "  {}: Passed".format(result_name) + Style.RESET_ALL)
        else:
            all_pass = False
            error_msgs[-1].append("Differences in {}\n{}".format(result_name, a - b))
            print(Fore.RED + "  {}: Failed".format(result_name) + Style.RESET_ALL)

    # iQMC flux
    if "iqmc" in output.keys():
        for score in [key for key in output["iqmc/tally/"].keys()]:
            result_name = "iqmc/tally/" + score
            a = output[result_name][:]
            b = answer[result_name][:]
            if a.size == 0:
                continue
            # Passed?
            if np.isclose(a, b).all():
                print(Fore.GREEN + "  {}: Passed".format(score) + Style.RESET_ALL)
            else:
                all_pass = False
                error_msgs[-1].append("Differences in {}\n{}".format(score, a - b))
                print(Fore.RED + "  {}: Failed".format(score) + Style.RESET_ALL)

    # Close files
    output.close()
    answer.close()

    # Move back up
    os.chdir("..")

# Report test results
N_fails = 0
for i in range(len(names)):
    if crashes[i] or len(error_msgs[i]) > 0:
        N_fails += 1

print(
    "\nTests passed: "
    + Fore.GREEN
    + "%i/%i" % (len(names) - N_fails, len(names))
    + Style.RESET_ALL
)
print("Tests failed: " + Fore.RED + "%i/%i" % (N_fails, len(names)) + Style.RESET_ALL)
print("  (%.2f seconds)\n" % np.sum(np.array(runtimes)))
for i in range(len(names)):
    if crashes[i]:
        print("\n" + "=" * 80)
        print("\n## {} crashed:".format(names[i]))
        print(printouts[i])
    if len(error_msgs[i]) > 0:
        print("\n" + "=" * 80)
        print("\n## {} failed:".format(names[i]))
        print(printouts[i])
        print("\n===\n")
        for msg in error_msgs[i]:
            print("\n# " + msg + "\n")

assert all_pass



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

import os, sys

sys.path.insert(0, os.path.abspath("../.."))


# On Read the Docs, need to mock any python packages that would require c
from unittest.mock import MagicMock

MOCK_MODULES = ["mpi4py"]
sys.modules.update((mod_name, MagicMock()) for mod_name in MOCK_MODULES)


# -- Project information -----------------------------------------------------

project = "MC/DC"
copyright = "2023, Center for Exascale Monte Carlo Neutron Transport (CEMeNT)"
author = "Center for Exascale Monte Carlo Neutron Transport (CEMeNT)"

# The full version, including alpha/beta/rc tags
release = " "


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx_toolbox.github",
    "sphinx_toolbox.sidebar_links",
]
autosummary_generate = True


github_username = "CEMeNT-PSAAP"
github_repository = "MCDC"
github_url = "https://github.com/{github_username}/{github_repository}"

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "furo"
html_logo = "images/home/mcdc.svg"

# html_permalinks = ['https://cement-psaap.github.io/', 'https://github.com/CEMeNT-PSAAP/MCDC']

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]



import numpy as np
from scipy.linalg import expm
from scipy.integrate import trapz

# Time grid
t = np.insert(np.logspace(-8, 1, 100), 0, 0.0)
K = len(t) - 1

# Load material data
with np.load("SHEM-361.npz") as data:
    SigmaT = data["SigmaT"]
    SigmaC = data["SigmaC"]
    SigmaS = data["SigmaS"]
    nuSigmaF_p = data["nuSigmaF_p"]
    SigmaF = data["SigmaF"]
    nu_p = data["nu_p"]
    nu_d = data["nu_d"]
    chi_p = data["chi_p"]
    chi_d = data["chi_d"]
    G = data["G"]
    J = data["J"]
    E = data["E"]
    v = data["v"]
    lamd = data["lamd"]
SigmaT += SigmaC * 0.28

# Matrix and RHS source
A = np.zeros([G + J, G + J])

# Top-left [GxG]: phi --> phi
A[:G, :G] = SigmaS + nuSigmaF_p - np.diag(SigmaT)

# Top-right [GxJ]: C --> phi
A[:G, G:] = np.multiply(chi_d, lamd)

# Bottom-left [JxG]: phi --> C
A[G:, :G] = np.multiply(nu_d, SigmaF)

# bottom-right [JxJ]: C --> C
A[G:, G:] = -np.diag(lamd)

# Multiply with neutron speed
AV = np.copy(A)
AV[:G, :] = np.dot(np.diag(v), A[:G, :])

# Initial condition
PHI_init = np.zeros(G + J)
PHI_init[G - 1] = v[-1]

# Allocate solution
PHI = np.zeros([K, G + J])

# Analytical solution
PHI_OLD = PHI_init
for k in range(K):
    print(k)
    t1 = t[k]
    t2 = t[k + 1]
    dt = t2 - t1

    # Fluxes
    PHI_NEW = np.dot(expm(AV * dt), PHI_OLD)
    PHI[k, :] = np.dot(np.linalg.inv(AV), (PHI_NEW - PHI_OLD) / dt)
    PHI_OLD[:] = PHI_NEW[:]
phi = PHI[:, :G]

# Density
n = np.zeros(K)
for k in range(K):
    n[k] = np.sum(phi[k, :] / v)

np.savez("reference.npz", t=t, phi=phi, n=n)



import numpy as np

from mcdc.type_ import nuclide
from mcdc.input_ import nuclide


def generate_test_nuclear_data(n_neutron_groups, n_dnp_groups):
    # Set random seed
    np.random.seed(90053)

    # Generate nuclear data
    capture = np.random.rand(n_neutron_groups)
    scatter = np.random.rand(n_neutron_groups, n_neutron_groups)
    fission = np.random.rand(n_neutron_groups)
    nu_s = np.random.rand(n_neutron_groups)
    nu_p = np.random.rand(n_neutron_groups)
    nu_d = np.random.rand(n_dnp_groups, n_neutron_groups)
    chi_p = np.random.rand(n_neutron_groups, n_neutron_groups)
    chi_d = np.random.rand(n_neutron_groups, n_dnp_groups)
    speed = np.random.rand(n_neutron_groups)
    decay = np.random.rand(n_dnp_groups)

    return [capture, scatter, fission, nu_s, nu_p, nu_d, chi_p, chi_d, speed, decay]


def verify_card_quantities_with_input_quantities(
    mat, capture, scatter, fission, nu_s, nu_p, nu_d, chi_p, chi_d, speed, decay
):
    # Verify pass-through quantities are correct
    assert np.allclose(mat["capture"], capture)
    assert np.allclose(mat["fission"], fission)
    assert np.allclose(mat["nu_p"], nu_p)
    assert np.allclose(mat["nu_s"], nu_s)
    assert np.allclose(mat["speed"], speed)
    assert np.allclose(mat["decay"], decay)

    # Verify modified and/or calculated quantities are correct
    n_neutron_groups = len(capture)
    n_dnp_groups = len(decay)
    total_scatter = np.sum(scatter, 0)
    total = capture + total_scatter + fission
    nu_f = nu_p + np.sum(np.transpose(nu_d), 1)
    normalized_chi_s = scatter * np.divide(1.0, np.sum(scatter, 0))
    normalized_chi_p = chi_p * np.divide(1.0, np.sum(chi_p, 0))
    normalized_chi_d = chi_d * np.divide(1.0, np.sum(chi_d, 0))
    assert mat["G"] == n_neutron_groups
    assert mat["J"] == n_dnp_groups
    assert np.allclose(mat["scatter"], total_scatter)
    assert np.allclose(mat["total"], total)
    assert np.allclose(mat["nu_d"], np.transpose(nu_d))
    assert np.allclose(mat["nu_f"], nu_f)
    assert np.allclose(mat["chi_s"], np.transpose(normalized_chi_s))
    assert np.allclose(mat["chi_p"], np.transpose(normalized_chi_p))
    assert np.allclose(mat["chi_d"], np.transpose(normalized_chi_d))


def test_single_neutron_energy_and_zero_dnp_group_nuclide():
    n_neutron_groups = 1
    n_dnp_groups = 0

    # Generate data for one energy group and one DNP group
    [
        capture,
        scatter,
        fission,
        nu_s,
        nu_p,
        nu_d,
        chi_p,
        chi_d,
        speed,
        decay,
    ] = generate_test_nuclear_data(n_neutron_groups, n_dnp_groups)

    # Initialize nuclide with single energy and DNP group
    m = nuclide(
        capture=capture,
        scatter=scatter,
        fission=fission,
        nu_p=nu_p,
        nu_d=nu_d,
        chi_p=chi_p,
        chi_d=chi_d,
        nu_s=nu_s,
        speed=speed,
        decay=decay,
    )

    # Verify input processing
    verify_card_quantities_with_input_quantities(
        m, capture, scatter, fission, nu_s, nu_p, nu_d, chi_p, chi_d, speed, decay
    )


def test_single_neutron_energy_and_single_dnp_group_nuclide():
    n_neutron_groups = 1
    n_dnp_groups = 1

    # Generate data for one energy group and one DNP group
    [
        capture,
        scatter,
        fission,
        nu_s,
        nu_p,
        nu_d,
        chi_p,
        chi_d,
        speed,
        decay,
    ] = generate_test_nuclear_data(n_neutron_groups, n_dnp_groups)

    # Initialize nuclide with single energy and DNP group
    m = nuclide(
        capture=capture,
        scatter=scatter,
        fission=fission,
        nu_p=nu_p,
        nu_d=nu_d,
        chi_p=chi_p,
        chi_d=chi_d,
        nu_s=nu_s,
        speed=speed,
        decay=decay,
    )

    # Verify input processing
    verify_card_quantities_with_input_quantities(
        m, capture, scatter, fission, nu_s, nu_p, nu_d, chi_p, chi_d, speed, decay
    )


def test_multiple_neutron_energy_and_zero_dnp_group_nuclide():
    n_neutron_groups = 7
    n_dnp_groups = 0

    # Generate data for one energy group and one DNP group
    [
        capture,
        scatter,
        fission,
        nu_s,
        nu_p,
        nu_d,
        chi_p,
        chi_d,
        speed,
        decay,
    ] = generate_test_nuclear_data(n_neutron_groups, n_dnp_groups)

    # Initialize nuclide with single energy and DNP group
    m = nuclide(
        capture=capture,
        scatter=scatter,
        fission=fission,
        nu_p=nu_p,
        nu_d=nu_d,
        chi_p=chi_p,
        chi_d=chi_d,
        nu_s=nu_s,
        speed=speed,
        decay=decay,
    )

    # Verify input processing
    verify_card_quantities_with_input_quantities(
        m, capture, scatter, fission, nu_s, nu_p, nu_d, chi_p, chi_d, speed, decay
    )


def test_multiple_neutron_energy_and_single_dnp_group_nuclide():
    n_neutron_groups = 7
    n_dnp_groups = 1

    # Generate data for one energy group and one DNP group
    [
        capture,
        scatter,
        fission,
        nu_s,
        nu_p,
        nu_d,
        chi_p,
        chi_d,
        speed,
        decay,
    ] = generate_test_nuclear_data(n_neutron_groups, n_dnp_groups)

    # Initialize nuclide with single energy and DNP group
    m = nuclide(
        capture=capture,
        scatter=scatter,
        fission=fission,
        nu_p=nu_p,
        nu_d=nu_d,
        chi_p=chi_p,
        chi_d=chi_d,
        nu_s=nu_s,
        speed=speed,
        decay=decay,
    )

    # Verify input processing
    verify_card_quantities_with_input_quantities(
        m, capture, scatter, fission, nu_s, nu_p, nu_d, chi_p, chi_d, speed, decay
    )


def test_single_neutron_energy_and_multiple_dnp_group_nuclide():
    n_neutron_groups = 1
    n_dnp_groups = 6

    # Generate data for one energy group and one DNP group
    [
        capture,
        scatter,
        fission,
        nu_s,
        nu_p,
        nu_d,
        chi_p,
        chi_d,
        speed,
        decay,
    ] = generate_test_nuclear_data(n_neutron_groups, n_dnp_groups)

    # Initialize nuclide with single energy and DNP group
    m = nuclide(
        capture=capture,
        scatter=scatter,
        fission=fission,
        nu_p=nu_p,
        nu_d=nu_d,
        chi_p=chi_p,
        chi_d=chi_d,
        nu_s=nu_s,
        speed=speed,
        decay=decay,
    )

    # Verify input processing
    verify_card_quantities_with_input_quantities(
        m, capture, scatter, fission, nu_s, nu_p, nu_d, chi_p, chi_d, speed, decay
    )


def test_multiple_neutron_energy_and_multiple_dnp_group_nuclide():
    n_neutron_groups = 7
    n_dnp_groups = 6

    # Generate data for one energy group and one DNP group
    [
        capture,
        scatter,
        fission,
        nu_s,
        nu_p,
        nu_d,
        chi_p,
        chi_d,
        speed,
        decay,
    ] = generate_test_nuclear_data(n_neutron_groups, n_dnp_groups)

    # Initialize nuclide with single energy and DNP group
    m = nuclide(
        capture=capture,
        scatter=scatter,
        fission=fission,
        nu_p=nu_p,
        nu_d=nu_d,
        chi_p=chi_p,
        chi_d=chi_d,
        nu_s=nu_s,
        speed=speed,
        decay=decay,
    )

    # Verify input processing
    verify_card_quantities_with_input_quantities(
        m, capture, scatter, fission, nu_s, nu_p, nu_d, chi_p, chi_d, speed, decay
    )



from reference import reference
import numpy as np
import h5py
import sys

sys.path.append("../../")
import tool


# Cases run
N_min = int(sys.argv[1])
N_max = int(sys.argv[2])
N_particle_list = np.logspace(N_min, N_max, (N_max - N_min) * 2 + 1)

# Reference solution
with h5py.File("output_%i.h5" % int(N_particle_list[0]), "r") as f:
    z = f["tally/grid/z"][:]
    mu = f["tally/grid/mu"][:]
phi_ref, J_ref, psi_ref = reference(z, mu)

# Error containers
error = np.zeros(len(N_particle_list))
error_J = np.zeros(len(N_particle_list))
error_psi = np.zeros(len(N_particle_list))

error_max = np.zeros(len(N_particle_list))
error_max_J = np.zeros(len(N_particle_list))
error_max_psi = np.zeros(len(N_particle_list))

# Calculate error
for k, N_particle in enumerate(N_particle_list):
    # Get results
    with h5py.File("output_%i.h5" % int(N_particle), "r") as f:
        z = f["tally/grid/z"][:]
        dz = z[1:] - z[:-1]
        mu = f["tally/grid/mu"][:]
        dmu = mu[1:] - mu[:-1]
        I = len(z) - 1
        N = len(mu) - 1

        psi = f["tally/flux/mean"][:]
        J = f["tally/current/mean"][:, 2]

    # Scalar flux
    phi = np.zeros(I)
    for i in range(I):
        phi[i] += np.sum(psi[i, :])

    psi_norm = np.zeros(psi.shape)
    # Normalize
    phi /= dz
    J /= dz
    for n in range(N):
        psi[:, n] = psi[:, n] / dz / dmu[n]

    # Get error
    error[k] = tool.rerror(phi, phi_ref)
    error_J[k] = tool.rerror(J, J_ref)
    error_psi[k] = tool.rerror(psi, psi_ref)

    error_max[k] = tool.rerror_max(phi, phi_ref)
    error_max_J[k] = tool.rerror_max(J, J_ref)
    error_max_psi[k] = tool.rerror_max(psi, psi_ref)


# Plot
tool.plot_convergence("slab_absorbium_flux", N_particle_list, error, error_max)
tool.plot_convergence("slab_absorbium_current", N_particle_list, error_J, error_max_J)
tool.plot_convergence(
    "slab_absorbium_angular_flux", N_particle_list, error_psi, error_max_psi
)



import numpy as np
import h5py

import mcdc


# =========================================================================
# Set model and run
# =========================================================================

n1 = mcdc.nuclide(capture=np.array([0.5]))
n2 = mcdc.nuclide(
    capture=np.array([0.1]),
    fission=np.array([0.4]),
    nu_p=np.array([2.5]),
    sensitivity=True,
)

m = mcdc.material(nuclides=[(n1, 1.0), (n2, 1.0)])

s1 = mcdc.surface("plane-x", x=-1e10, bc="reflective")
s2 = mcdc.surface("plane-x", x=1e10, bc="reflective")

mcdc.cell([+s1, -s2], m)

mcdc.source(point=[0.0, 0.0, 0.0], isotropic=True)

scores = ["flux"]
mcdc.tally(
    scores=scores,
    x=np.linspace(-20.0, 20.0, 202),
    t=np.linspace(0.0, 20.0, 21),
)

mcdc.setting(N_particle=30)

mcdc.run()



import numpy as np
from scipy.integrate import quad
from scipy.special import exp1

# Parameters
SigmaT = 1.0
v = 1.0
T = 5.0

# Point-wise solution


def phi_(x, t):
    if x > v * t:
        return 0.0
    else:
        return (
            1.0
            / T
            * (
                SigmaT * x * (exp1(SigmaT * v * t) - exp1(SigmaT * x))
                + np.e ** (-SigmaT * x)
                - x / (v * t) * np.e ** (-SigmaT * v * t)
            )
        )


def phi_X(t, x1, x2):
    return quad(phi_, x1, x2, args=(t))[0] / (x2 - x1)


def reference(x, t):
    dx = x[1] - x[0]
    x_mid = 0.5 * (x[:-1] + x[1:])
    dt = t[1:] - t[:-1]
    K = len(dt)
    J = len(x_mid)

    phi = np.zeros([K, J])
    for k in range(K):
        for j in range(J):
            x1 = x[j]
            x2 = x[j + 1]
            t1 = t[k]
            t2 = t[k + 1]
            phi[k, j] = quad(phi_X, t1, t2, args=(x1, x2))[0] / (t2 - t1)
            # phi[k, j] = quad(phi_, x1, x2, args=(t[k + 1]))[0] / (x2 - x1)

    return phi



import numpy as np
import h5py

import mcdc


# =========================================================================
# Set model and run
# =========================================================================

lib = h5py.File("c5g7.h5", "r")


def set_mat(mat):
    return mcdc.material(
        capture=mat["capture"][:],
        scatter=mat["scatter"][:],
        fission=mat["fission"][:],
        nu_p=mat["nu_p"][:],
        nu_d=mat["nu_d"][:],
        chi_p=mat["chi_p"][:],
        chi_d=mat["chi_d"][:],
        speed=mat["speed"],
        decay=mat["decay"],
        sensitivity=True,
    )


mat_uo2 = set_mat(lib["uo2"])  # Fuel: UO2
mat_mod = set_mat(lib["mod"])  # Moderator
mat_cr = set_mat(lib["cr"])  # Control rod

s1 = mcdc.surface("plane-x", x=0.0, bc="reflective")
s2 = mcdc.surface("plane-x", x=0.5, sensitivity=True)
s3 = mcdc.surface("plane-x", x=1.5, sensitivity=True)
s4 = mcdc.surface("plane-x", x=2.0, bc="reflective")

mcdc.cell([+s1, -s2], mat_uo2)
mcdc.cell([+s2, -s3], mat_mod)
mcdc.cell([+s3, -s4], mat_cr)

mcdc.source(point=[1.0, 0.0, 0.0], energy=[1, 0, 0, 0, 0, 0, 0], isotropic=True)

scores = ["flux"]
mcdc.tally(scores=scores, x=np.linspace(0.0, 2.0, 11), g="all")

mcdc.setting(N_particle=3)

mcdc.run()



import matplotlib.pyplot as plt
import h5py
import numpy as np
import sys

from reference import reference


# Reference solution
output = sys.argv[1]
with h5py.File(output, "r") as f:
    z = f["tally/grid/z"][:]
    mu = f["tally/grid/mu"][:]
phi_ref, J_ref, psi_ref = reference(z, mu)

# Load results
with h5py.File(output, "r") as f:
    z = f["tally/grid/z"][:]
    dz = z[1:] - z[:-1]
    z_mid = 0.5 * (z[:-1] + z[1:])
    mu = f["tally/grid/mu"][:]
    dmu = mu[1:] - mu[:-1]
    mu_mid = 0.5 * (mu[:-1] + mu[1:])
    I = len(z) - 1
    N = len(mu) - 1

    psi = f["tally/flux/mean"][:]
    psi_sd = f["tally/flux/sdev"][:]
    J = f["tally/current/mean"][:, 2]
    J_sd = f["tally/current/sdev"][:, 2]

# Scalar flux
phi = np.zeros(I)
phi_sd = np.zeros(I)
for i in range(I):
    phi[i] += np.sum(psi[i, :])
    phi_sd[i] += np.linalg.norm(psi_sd[i, :])

# Normalize
phi /= dz
phi_sd /= dz
J /= dz
J_sd /= dz
for n in range(N):
    psi[:, n] = psi[:, n] / dz / dmu[n]
    psi_sd[:, n] = psi_sd[:, n] / dz / dmu[n]

# Flux - spatial average
plt.plot(z_mid, phi, "-b", label="MC")
plt.fill_between(z_mid, phi - phi_sd, phi + phi_sd, alpha=0.2, color="b")
plt.plot(z_mid, phi_ref, "--r", label="Ref.")
plt.xlabel(r"$z$, cm")
plt.ylabel("Flux")
plt.ylim([0.06, 0.16])
plt.grid()
plt.legend()
plt.title(r"$\bar{\phi}_i$")
plt.show()

# Current - spatial average
plt.plot(z_mid, J, "-b", label="MC")
plt.fill_between(z_mid, J - J_sd, J + J_sd, alpha=0.2, color="b")
plt.plot(z_mid, J_ref, "--r", label="Ref.")
plt.xlabel(r"$z$, cm")
plt.ylabel("Current")
plt.ylim([-0.03, 0.045])
plt.grid()
plt.legend()
plt.title(r"$\bar{J}_i$")
plt.show()

# Angular flux - spatial average
vmin = min(np.min(psi_ref), np.min(psi))
vmax = max(np.max(psi_ref), np.max(psi))
fig, ax = plt.subplots(1, 2, figsize=(4, 3), sharey=True)
Z, MU = np.meshgrid(z_mid, mu_mid)
im = ax[0].pcolormesh(MU.T, Z.T, psi_ref, vmin=vmin, vmax=vmax)
ax[0].set_xlabel(r"Polar cosine, $\mu_n$")
ax[0].set_ylabel(r"$z$")
ax[0].set_title(r"$\bar{\psi}_i(\mu_n)$ [Ref.]")
ax[1].pcolormesh(MU.T, Z.T, psi, vmin=vmin, vmax=vmax)
ax[1].set_xlabel(r"Polar cosine, $\mu_n$")
ax[1].set_ylabel(r"$z$")
ax[1].set_title(r"$\bar{\psi}_i(\mu_n)$ [MC]")
fig.subplots_adjust(right=0.8)
cbar_ax = fig.add_axes([0.85, 0.15, 0.05, 0.7])
cbar = fig.colorbar(im, cax=cbar_ax)
cbar.set_label("Angular flux")
plt.show()
